########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import asyncio
import contextlib
import itertools
import logging
import unittest

import aioxmpp.callbacks as callbacks
import aioxmpp.service as service
import aioxmpp.stream
import aioxmpp.xso

from aioxmpp.testutils import (
    run_coroutine,
    CoroutineMock,
)


class TestServiceMeta(unittest.TestCase):
    def setUp(self):
        self.descriptor_cm = unittest.mock.Mock()

        class FooDescriptor(service.Descriptor):
            def __init__(self, *, deps=[]):
                super().__init__()
                self._deps = list(deps)

            @property
            def required_dependencies(self):
                return list(self._deps)

            def init_cm(self, instance):
                return self.descriptor_cm(instance)

            @property
            def value_type(self):
                return None

        self.FooDescriptor = FooDescriptor

    def tearDown(self):
        del self.FooDescriptor

    def test_inherits_from_ABCMeta(self):
        self.assertTrue(issubclass(service.Meta, abc.ABCMeta))

    def test_ordering_attributes(self):
        class Foo(metaclass=service.Meta):
            pass

        self.assertSetEqual(
            set(),
            Foo.ORDER_BEFORE
        )
        self.assertSetEqual(
            set(),
            Foo.ORDER_AFTER
        )

    def test_defining_PATCHED_ORDER_AFTER_raises(self):

        with self.assertRaisesRegex(
                TypeError,
                "PATCHED_ORDER_AFTER must not be defined manually\. "
                "it is supplied automatically by the metaclass\."):
            class Bar(metaclass=service.Meta):
                pass

            class Foo(metaclass=service.Meta):
                PATCHED_ORDER_AFTER = [Bar]

    def test_defining_DEPGRAPH_NODE_raises(self):

        with self.assertRaisesRegex(
                TypeError,
                "_DEPGRAPH_NODE must not be defined manually\. "
                "it is supplied automatically by the metaclass\."):

            class Foo(metaclass=service.Meta):
                _DEPGRAPH_NODE = None

    def test_configure_ordering(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_BEFORE = [Foo]

        self.assertSetEqual(
            {Foo},
            Bar.ORDER_BEFORE
        )
        self.assertSetEqual(
            set(),
            Bar.ORDER_AFTER
        )
        self.assertSetEqual(
            set(),
            Bar.PATCHED_ORDER_AFTER
        )
        self.assertSetEqual(
            set(),
            Foo.ORDER_AFTER
        )
        self.assertSetEqual(
            set(),
            Foo.ORDER_BEFORE
        )
        self.assertSetEqual(
            {Bar},
            Foo.PATCHED_ORDER_AFTER
        )

    def test_transitive_before_ordering(self):

        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_BEFORE = [Foo]

        class Baz(metaclass=service.Meta):
            ORDER_BEFORE = [Bar]

        self.assertGreater(Foo, Bar)
        self.assertGreater(Bar, Baz)
        self.assertGreater(Foo, Baz)

    def test_transitive_after_ordering(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_AFTER = [Foo]

        class Baz(metaclass=service.Meta):
            ORDER_AFTER = [Bar]

        self.assertLess(Foo, Bar)
        self.assertLess(Bar, Baz)
        self.assertLess(Foo, Baz)

    def test_loop_detect(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_AFTER = [Foo]

        with self.assertRaisesRegex(
                ValueError,
                "dependency loop in service definitions"):

            class Fnord(metaclass=service.Meta):
                ORDER_BEFORE = [Foo]
                ORDER_AFTER = [Bar]

            print(Fnord.ORDER_BEFORE)
            print(Fnord.ORDER_AFTER)

    def test_topological_dependency_ordering_puts_earliest_first(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_BEFORE = [Foo]

        class Baz(metaclass=service.Meta):
            ORDER_BEFORE = [Bar]

        class Fourth(metaclass=service.Meta):
            ORDER_BEFORE = [Bar]

        self.assertLess(Baz, Bar)
        self.assertLess(Fourth, Bar)
        self.assertLess(Bar, Foo)
        self.assertLess(Baz, Foo)
        self.assertLess(Fourth, Foo)

        self.assertLessEqual(Baz, Bar)
        self.assertLessEqual(Fourth, Bar)
        self.assertLessEqual(Bar, Foo)
        self.assertLessEqual(Baz, Foo)
        self.assertLessEqual(Fourth, Foo)

        self.assertGreater(Foo, Bar)
        self.assertGreater(Foo, Baz)
        self.assertGreater(Foo, Fourth)
        self.assertGreater(Bar, Baz)
        self.assertGreater(Bar, Fourth)

        self.assertGreaterEqual(Foo, Bar)
        self.assertGreaterEqual(Foo, Baz)
        self.assertGreaterEqual(Foo, Fourth)
        self.assertGreaterEqual(Bar, Baz)
        self.assertGreaterEqual(Bar, Fourth)

    def test_topological_ordering_fail2(self):
        class A(metaclass=service.Meta):
            pass

        self.assertEqual(A, A)
        self.assertLessEqual(A, A)
        self.assertGreaterEqual(A, A)

    def test_topological_ordering_sort_fail(self):
        class A(metaclass=service.Meta):
            pass

        class B(metaclass=service.Meta):
            pass

        class C(metaclass=service.Meta):
            ORDER_AFTER = [B]

        class D(metaclass=service.Meta):
            ORDER_BEFORE = [A, C]

        for services in itertools.permutations([A, B, C, D]):
            services = list(services)
            services.sort()

            if services.index(C) < services.index(B):
                self.fail(services)

    def test_inheritance_ignores_non_service_classes(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar:
            ORDER_BEFORE = [Foo]

        class Baz(Bar, metaclass=service.Meta):
            pass

        self.assertSetEqual(set(), Baz.ORDER_BEFORE)

    def test_support_pre_0_3_attributes_on_class(self):
        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            pass

        class A(metaclass=service.Meta):
            SERVICE_BEFORE = [Foo]
            SERVICE_AFTER = [Bar]

        self.assertGreater(Foo, A)
        self.assertGreater(A, Bar)
        self.assertGreater(Foo, Bar)

        self.assertEqual(
            Foo.PATCHED_ORDER_AFTER,
            {A}
        )

    def test_support_pre_0_3_attributes_on_read(self):
        class Foo(metaclass=service.Meta):
            pass

        self.assertIs(Foo.ORDER_BEFORE, Foo.SERVICE_BEFORE)
        self.assertIs(Foo.ORDER_AFTER, Foo.SERVICE_AFTER)

    def test_support_pre_0_3_attributes_raise_if_both_are_given(self):
        class Foo(metaclass=service.Meta):
            pass

        with self.assertRaisesRegex(ValueError, "mixes old and new"):
            class Bar1(metaclass=service.Meta):
                ORDER_BEFORE = [Foo]
                SERVICE_BEFORE = [Foo]

        with self.assertRaisesRegex(ValueError, "mixes old and new"):
            class Bar2(metaclass=service.Meta):
                ORDER_AFTER = [Foo]
                SERVICE_AFTER = [Foo]

        with self.assertRaisesRegex(ValueError, "mixes old and new"):
            class Bar3(metaclass=service.Meta):
                ORDER_BEFORE = [Foo]
                SERVICE_AFTER = [Foo]

    def test_support_pre_0_3_attributes_with_deprecation_warning(self):
        class Foo(metaclass=service.Meta):
            pass

        with unittest.mock.patch("warnings.warn") as warn:
            class Bar1(metaclass=service.Meta):
                SERVICE_BEFORE = [Foo]

            class Bar2(metaclass=service.Meta):
                SERVICE_AFTER = [Foo]

        s = "SERVICE_BEFORE/AFTER used on class; use ORDER_BEFORE/AFTER"
        self.assertSequenceEqual(
            warn.mock_calls,
            [
                unittest.mock.call(s, DeprecationWarning),
                unittest.mock.call(s, DeprecationWarning),
            ]
        )

    def test_collect_handlers(self):
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        class Foo(metaclass=service.Meta):
            x = ObjectWithHandlers(
                [
                    service.HandlerSpec(
                        unittest.mock.sentinel.k1,
                        is_unique=True,
                    ),
                    service.HandlerSpec(
                        unittest.mock.sentinel.k2,
                        is_unique=False,
                    ),
                ]
            )

        self.assertCountEqual(
            Foo.SERVICE_HANDLERS,
            (
                (unittest.mock.sentinel.k1, Foo.x),
                (unittest.mock.sentinel.k2, Foo.x),
            ),
        )

        self.assertIsInstance(Foo.SERVICE_HANDLERS, tuple)

    def test_reject_missing_dependencies(self):
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        with self.assertRaisesRegex(
                TypeError,
                r"decorator requires dependency .* but it is not declared"):
            class Foo(metaclass=service.Meta):
                x = ObjectWithHandlers(
                    [
                        service.HandlerSpec(
                            unittest.mock.sentinel.k1,
                            is_unique=True,
                            require_deps=(unittest.mock.sentinel.nonexistent,)
                        ),
                    ]
                )

    def test_allow_properly_declared_dependencies(self):
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        class Other(metaclass=service.Meta):
            pass

        class Foo(metaclass=service.Meta):
            ORDER_AFTER = [Other]

            x = ObjectWithHandlers(
                [
                    service.HandlerSpec(
                        unittest.mock.sentinel.k1,
                        is_unique=True,
                        require_deps=(Other,)
                    ),
                ]
            )

    def test_reject_duplicate_handlers_on_different_objects(self):
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        with self.assertRaisesRegex(
                TypeError,
                "handler conflict between .* and .*: both want to use .*"):

            class Foo(metaclass=service.Meta):
                x = ObjectWithHandlers(
                    [
                        service.HandlerSpec(
                            unittest.mock.sentinel.k1
                        ),
                        service.HandlerSpec(
                            unittest.mock.sentinel.k2
                        ),
                    ]
                )

                y = ObjectWithHandlers(
                    [
                        service.HandlerSpec(
                            unittest.mock.sentinel.k2
                        ),
                        service.HandlerSpec(
                            unittest.mock.sentinel.k3
                        )
                    ]
                )

    def test_allow_duplicate_handlers_on_different_objects_for_non_unique(self):  # NOQA
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        class Foo(metaclass=service.Meta):
            x = ObjectWithHandlers(
                [
                    service.HandlerSpec(
                        unittest.mock.sentinel.k1
                    ),
                    service.HandlerSpec(
                        unittest.mock.sentinel.k2,
                        is_unique=False,
                    ),
                ]
            )

            y = ObjectWithHandlers(
                [
                    service.HandlerSpec(
                        unittest.mock.sentinel.k2,
                        is_unique=False,
                    ),
                    service.HandlerSpec(
                        unittest.mock.sentinel.k3
                    )
                ]
            )

    def test_reject_inheritance_from_class_with_handlers(self):
        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

        class Foo(metaclass=service.Meta):
            x = ObjectWithHandlers(
                [
                    service.HandlerSpec(
                        unittest.mock.sentinel.k1
                    ),
                ]
            )

        with self.assertRaisesRegex(
                TypeError,
                r"subclassing services is prohibited."):
            class Bar(Foo):
                pass

    def test_collect_descriptors(self):
        descriptor = self.FooDescriptor()

        class Foo(metaclass=service.Meta):
            x = descriptor

        self.assertCountEqual(
            Foo.SERVICE_HANDLERS,
            (
                descriptor,
            ),
        )

    def test_reject_inheritance_from_class_with_descriptors(self):
        class Foo(metaclass=service.Meta):
            x = self.FooDescriptor()

        with self.assertRaisesRegex(
                TypeError,
                r"subclassing services is prohibited."):
            class Bar(Foo):
                pass

    def test_reject_missing_descriptor_dependencies(self):
        with self.assertRaisesRegex(
                TypeError,
                r"descriptor requires dependency .* but it is not declared"):
            class Foo(metaclass=service.Meta):
                x = self.FooDescriptor(deps=[unittest.mock.sentinel.foo])

    def test_allow_properly_declared_descriptor_dependencies(self):
        class Other(metaclass=service.Meta):
            pass

        class Foo(metaclass=service.Meta):
            ORDER_AFTER = [Other]

            x = self.FooDescriptor(deps=[Other])


class TestService(unittest.TestCase):
    def test_is_Meta(self):
        self.assertIsInstance(
            service.Service,
            service.Meta
        )

    def test_automatic_logger(self):
        class Service(service.Service):
            pass

        s = Service(None)
        self.assertIsInstance(s.logger, logging.Logger)
        self.assertEqual(
            "tests.test_service.TestService"
            ".test_automatic_logger.<locals>.Service",
            s.logger.name
        )

    def test_custom_logger(self):
        l = logging.getLogger("foo")
        s = service.Service(None, logger_base=l)

        self.assertEqual(s.logger, l.getChild("service.Service"))

    def test_client(self):
        o = object()
        s = service.Service(o)
        self.assertIs(s.client, o)

        with self.assertRaises(AttributeError):
            s.client = o

    def test_shutdown(self):
        o = object()
        s = service.Service(o)

        def coro():
            return
            yield

        with unittest.mock.patch.object(s, "_shutdown") as _shutdown:
            _shutdown.return_value = coro()
            run_coroutine(s.shutdown())

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            _shutdown.mock_calls
        )

    def test_setup_handler_context(self):
        res1 = unittest.mock.Mock()
        res2 = unittest.mock.Mock()

        base = unittest.mock.Mock()

        self.maxDiff = None

        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

            def __get__(self, instance, owner):
                if instance is None:
                    return self
                return (unittest.mock.sentinel.got, self)

        class ServiceWithHandlers(service.Service):
            x = ObjectWithHandlers(
                [
                    service.HandlerSpec((res1, ()), is_unique=False),
                    service.HandlerSpec((res2, ("foo",))),
                ]
            )

            y = ObjectWithHandlers(
                [
                    service.HandlerSpec((res2, ("bar",))),
                ]
            )

        client = unittest.mock.Mock()

        with unittest.mock.patch("contextlib.ExitStack",
                                 new=base.ExitStack) as ExitStack:
            s = ServiceWithHandlers(client)

        self.assertCountEqual(
            res1.mock_calls,
            [
                unittest.mock.call(
                    s,
                    client.stream,
                    (unittest.mock.sentinel.got, ServiceWithHandlers.x)
                )
            ]
        )

        self.assertCountEqual(
            res2.mock_calls,
            [
                unittest.mock.call(
                    s,
                    client.stream,
                    (unittest.mock.sentinel.got, ServiceWithHandlers.x),
                    "foo",
                ),
                unittest.mock.call(
                    s,
                    client.stream,
                    (unittest.mock.sentinel.got, ServiceWithHandlers.y),
                    "bar",
                )
            ]
        )

        self.assertSequenceEqual(
            ExitStack.mock_calls,
            [
                unittest.mock.call(),
            ] + [
                unittest.mock.call().enter_context(
                    handler_cm()
                )
                for (handler_cm, args), obj
                in ServiceWithHandlers.SERVICE_HANDLERS
            ]
        )

        base.mock_calls.clear()

        base._shutdown = CoroutineMock()

        with unittest.mock.patch.object(s, "_shutdown", new=base._shutdown):
            run_coroutine(s.shutdown())

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call._shutdown(),
                unittest.mock.call.ExitStack().close(),
            ]
        )

    def test_setup_descriptor_context(self):
        class FooDescriptor(service.Descriptor):
            def init_cm(self, instance):
                return base.init_cm(instance)

            def add_to_stack(self, instance, stack):
                base.add_to_stack(self, instance, stack)

            @property
            def value_type(self):
                return None

        base = unittest.mock.Mock()
        base.init_cm = unittest.mock.MagicMock()

        self.maxDiff = None

        class ServiceWithDescriptor(service.Service):
            desc = FooDescriptor()

        client = unittest.mock.Mock()

        with unittest.mock.patch("contextlib.ExitStack",
                                 new=base.ExitStack):
            s = ServiceWithDescriptor(client)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.ExitStack(),
                unittest.mock.call.add_to_stack(
                    ServiceWithDescriptor.desc,
                    s,
                    base.ExitStack(),
                )
            ]
        )

        base.mock_calls.clear()

        base._shutdown = CoroutineMock()

        with unittest.mock.patch.object(s, "_shutdown", new=base._shutdown):
            run_coroutine(s.shutdown())

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call._shutdown(),
                unittest.mock.call.ExitStack().close(),
            ]
        )

    def test_setup_contexts_in_delaration_order(self):
        class FooDescriptor(service.Descriptor):
            def __init__(self, id_):
                super().__init__()
                self.__id = id_

            def init_cm(self, instance):
                return base.init_cm(instance, self.__id)

            def add_to_stack(self, instance, stack):
                base.add_to_stack(self, instance, stack)

            @property
            def value_type(self):
                return None

        base = unittest.mock.Mock()
        base.init_cm = unittest.mock.MagicMock()

        self.maxDiff = None

        class ObjectWithHandlers:
            def __init__(self, handlers=[]):
                self._aioxmpp_service_handlers = set(handlers)

            def __get__(self, instance, owner):
                if instance is None:
                    return self
                return (unittest.mock.sentinel.got, self)

        class ServiceWithDescriptor(service.Service):
            desc1 = FooDescriptor(1)

            x = ObjectWithHandlers(
                [
                    service.HandlerSpec((base.res1, ())),
                ]
            )

            desc2 = FooDescriptor(2)

        client = unittest.mock.Mock()

        with unittest.mock.patch("contextlib.ExitStack",
                                 new=base.ExitStack):
            s = ServiceWithDescriptor(client)

        calls = list(base.mock_calls)
        self.assertSequenceEqual(
            calls,
            [
                unittest.mock.call.ExitStack(),
                unittest.mock.call.add_to_stack(
                    ServiceWithDescriptor.desc1,
                    s,
                    base.ExitStack(),
                ),
                unittest.mock.call.res1(
                    s,
                    client.stream,
                    (unittest.mock.sentinel.got, ServiceWithDescriptor.x)
                ),
                unittest.mock.call.ExitStack().enter_context(
                    base.res1(),
                ),
                unittest.mock.call.add_to_stack(
                    ServiceWithDescriptor.desc2,
                    s,
                    base.ExitStack(),
                )
            ]
        )

        base.mock_calls.clear()

        base._shutdown = CoroutineMock()

        with unittest.mock.patch.object(s, "_shutdown", new=base._shutdown):
            run_coroutine(s.shutdown())

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call._shutdown(),
                unittest.mock.call.ExitStack().close(),
            ]
        )

    def test_dependencies(self):
        s = service.Service(
            unittest.mock.sentinel.client,
            dependencies=unittest.mock.sentinel.deps
        )
        self.assertEqual(
            s.dependencies,
            unittest.mock.sentinel.deps,
        )

    def test_dependencies_is_not_writable(self):
        s = service.Service(
            unittest.mock.sentinel.client,
            dependencies=unittest.mock.sentinel.deps
        )
        with self.assertRaises(AttributeError):
            s.dependencies = unittest.mock.sentinel.deps


class Test_apply_iq_handler(unittest.TestCase):
    def test_uses_stream_iq_handler(self):
        with unittest.mock.patch("aioxmpp.stream.iq_handler") as iq_handler:
            service._apply_iq_handler(
                unittest.mock.sentinel.instance,
                unittest.mock.sentinel.stream,
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.payload_cls,
            )

        iq_handler.assert_called_with(
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.type_,
            unittest.mock.sentinel.payload_cls,
            unittest.mock.sentinel.func,
        )


class Test_apply_inbound_message_filter(unittest.TestCase):
    def test_uses_stream_stanza_filter(self):
        stream = unittest.mock.Mock()

        with unittest.mock.patch(
                "aioxmpp.stream.stanza_filter") as stanza_filter:
            service._apply_inbound_message_filter(
                unittest.mock.sentinel.instance,
                stream,
                unittest.mock.sentinel.func,
            )

        stanza_filter.assert_called_with(
            stream.service_inbound_message_filter,
            unittest.mock.sentinel.func,
            type(unittest.mock.sentinel.instance),
        )


class Test_apply_inbound_presence_filter(unittest.TestCase):
    def test_uses_stream_stanza_filter(self):
        stream = unittest.mock.Mock()

        with unittest.mock.patch(
                "aioxmpp.stream.stanza_filter") as stanza_filter:
            service._apply_inbound_presence_filter(
                unittest.mock.sentinel.instance,
                stream,
                unittest.mock.sentinel.func,
            )

        stanza_filter.assert_called_with(
            stream.service_inbound_presence_filter,
            unittest.mock.sentinel.func,
            type(unittest.mock.sentinel.instance),
        )


class Test_apply_outbound_message_filter(unittest.TestCase):
    def test_uses_stream_stanza_filter(self):
        stream = unittest.mock.Mock()

        with unittest.mock.patch(
                "aioxmpp.stream.stanza_filter") as stanza_filter:
            service._apply_outbound_message_filter(
                unittest.mock.sentinel.instance,
                stream,
                unittest.mock.sentinel.func,
            )

        stanza_filter.assert_called_with(
            stream.service_outbound_message_filter,
            unittest.mock.sentinel.func,
            type(unittest.mock.sentinel.instance),
        )


class Test_apply_outbound_presence_filter(unittest.TestCase):
    def test_uses_stream_stanza_filter(self):
        stream = unittest.mock.Mock()

        with unittest.mock.patch(
                "aioxmpp.stream.stanza_filter") as stanza_filter:
            service._apply_outbound_presence_filter(
                unittest.mock.sentinel.instance,
                stream,
                unittest.mock.sentinel.func,
            )

        stanza_filter.assert_called_with(
            stream.service_outbound_presence_filter,
            unittest.mock.sentinel.func,
            type(unittest.mock.sentinel.instance),
        )


class Test_apply_connect_depsignal(unittest.TestCase):
    def test_uses_dependencies(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency

        result = service._apply_connect_depsignal(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.dependency,
            "signal_name",
            unittest.mock.sentinel.mode,
        )

        instance.dependencies.__getitem__.assert_called_with(
            unittest.mock.sentinel.dependency,
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.signal_name.context_connect(
                    unittest.mock.sentinel.func,
                    unittest.mock.sentinel.mode,
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.signal_name.context_connect(),
        )

    def test_can_connect_to_StanzaStream(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.client.stream = dependency

        result = service._apply_connect_depsignal(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            aioxmpp.stream.StanzaStream,
            "signal_name",
            unittest.mock.sentinel.mode,
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.signal_name.context_connect(
                    unittest.mock.sentinel.func,
                    unittest.mock.sentinel.mode,
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.signal_name.context_connect(),
        )

    def test_can_connect_to_Client(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.client = dependency

        result = service._apply_connect_depsignal(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            aioxmpp.node.Client,
            "signal_name",
            unittest.mock.sentinel.mode,
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.signal_name.context_connect(
                    unittest.mock.sentinel.func,
                    unittest.mock.sentinel.mode,
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.signal_name.context_connect(),
        )

    def test_does_not_pass_mode_if_it_is_None(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency

        result = service._apply_connect_depsignal(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.dependency,
            "signal_name",
            None,
        )

        instance.dependencies.__getitem__.assert_called_with(
            unittest.mock.sentinel.dependency,
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.signal_name.context_connect(
                    unittest.mock.sentinel.func,
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.signal_name.context_connect(),
        )

    def test_runs_mode_if_tuple(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency
        mode_mock = unittest.mock.Mock()

        result = service._apply_connect_depsignal(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.dependency,
            "signal_name",
            (mode_mock, ("a", "b")),
        )

        instance.dependencies.__getitem__.assert_called_with(
            unittest.mock.sentinel.dependency,
        )

        mode_mock.assert_called_once_with("a", "b")

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.signal_name.context_connect(
                    unittest.mock.sentinel.func,
                    mode_mock(),
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.signal_name.context_connect(),
        )


class Test_apply_connect_depfilter(unittest.TestCase):
    def test_uses_dependencies(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.dependencies.__getitem__.return_value = dependency

        result = service._apply_connect_depfilter(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.dependency,
            "filter_name",
        )

        instance.dependencies.__getitem__.assert_called_with(
            unittest.mock.sentinel.dependency,
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.filter_name.context_register(
                    unittest.mock.sentinel.func,
                    type(instance),
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.filter_name.context_register(),
        )

    def test_can_connect_to_StanzaStream(self):
        instance = unittest.mock.MagicMock()
        dependency = unittest.mock.Mock()
        instance.client.stream = dependency

        result = service._apply_connect_depfilter(
            instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            aioxmpp.stream.StanzaStream,
            "filter_name",
        )

        self.assertSequenceEqual(
            dependency.mock_calls,
            [
                unittest.mock.call.filter_name.context_register(
                    unittest.mock.sentinel.func,
                    type(instance),
                )
            ]
        )

        self.assertEqual(
            result,
            dependency.filter_name.context_register(),
        )


class Test_apply_connect_attrsignal(unittest.TestCase):
    def test_uses_descriptor(self):
        descriptor = unittest.mock.PropertyMock()

        result = service._apply_connect_attrsignal(
            unittest.mock.sentinel.instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            descriptor,
            "signal_name",
            unittest.mock.sentinel.mode,
        )

        descriptor.assert_called_once_with()

        descriptor().signal_name.context_connect\
            .assert_called_once_with(
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.mode,
            )

        self.assertEqual(
            result,
            descriptor().signal_name.context_connect(),
        )

    def test_default_mode(self):
        descriptor = unittest.mock.PropertyMock()

        result = service._apply_connect_attrsignal(
            unittest.mock.sentinel.instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            descriptor,
            "signal_name",
            None,
        )

        descriptor.assert_called_once_with()

        descriptor().signal_name.context_connect\
            .assert_called_once_with(
                unittest.mock.sentinel.func,
            )

        self.assertEqual(
            result,
            descriptor().signal_name.context_connect(),
        )

    def test_runs_mode_if_tuple(self):
        mode_func = unittest.mock.Mock()
        descriptor = unittest.mock.PropertyMock()

        result = service._apply_connect_attrsignal(
            unittest.mock.sentinel.instance,
            unittest.mock.sentinel.stream,
            unittest.mock.sentinel.func,
            descriptor,
            "signal_name",
            (mode_func, (1, "foo")),
        )

        descriptor.assert_called_once_with()

        mode_func.assert_called_once_with(1, "foo")

        descriptor().signal_name.context_connect\
            .assert_called_once_with(
                unittest.mock.sentinel.func,
                mode_func(),
            )

        self.assertEqual(
            result,
            descriptor().signal_name.context_connect(),
        )


class Testadd_handler_spec(unittest.TestCase):
    def test_adds_magic_attribute(self):
        target = unittest.mock.Mock(spec=[])

        self.assertFalse(
            service.has_magic_attr(target),
        )

        service.add_handler_spec(
            target,
            unittest.mock.sentinel.foo,
        )

        self.assertTrue(
            service.has_magic_attr(target),
        )

        self.assertIn(
            unittest.mock.sentinel.foo,
            service.get_magic_attr(target),
        )

    def test_preserves_existing_magic_attr(self):
        target = unittest.mock.Mock(spec=[])
        service.automake_magic_attr(target)
        service.get_magic_attr(target).add(
            unittest.mock.sentinel.bar
        )

        self.assertTrue(
            service.has_magic_attr(target),
        )

        service.add_handler_spec(
            target,
            unittest.mock.sentinel.foo,
        )

        self.assertTrue(
            service.has_magic_attr(target),
        )

        self.assertCountEqual(
            service.get_magic_attr(target),
            [
                unittest.mock.sentinel.foo,
                unittest.mock.sentinel.bar,
            ]
        )


class Testiq_handler(unittest.TestCase):
    @aioxmpp.IQ.as_payload_class
    class Payload(aioxmpp.xso.XSO):
        TAG = ("urn:example:test_service:test_iq_handler", "valid-payload")

    def setUp(self):
        self.decorator = service.iq_handler(
            unittest.mock.sentinel.type_,
            self.Payload,
        )

    def tearDown(self):
        del self.decorator

    def test_rejects_non_registered_payload(self):
        class SomeXSO(aioxmpp.xso.XSO):
            TAG = ("urn:example:test_service:test_iq_handler", "invalid-payload")

        with self.assertRaisesRegex(
                ValueError,
                r".+ is not a valid IQ payload"):
            service.iq_handler(
                unittest.mock.sentinel.type_,
                SomeXSO,
            )

    def test_rejects_incomplete_xso_payload(self):
        class SomeXSO(aioxmpp.xso.XSO):
            pass

        with self.assertRaisesRegex(
                ValueError,
                r".+ is not a valid IQ payload"):
            service.iq_handler(
                unittest.mock.sentinel.type_,
                SomeXSO,
            )

    def test_rejects_incorrect_xso_payload(self):
        class SomeOtherXSO(aioxmpp.xso.XSO):
            TAG = ("urn:example:test_service:test_iq_handler", "valid-payload")

        with self.assertRaisesRegex(
                ValueError,
                r".+ is not a valid IQ payload"):
            service.iq_handler(
                unittest.mock.sentinel.type_,
                SomeOtherXSO,
            )

    def test_works_as_decorator(self):
        @asyncio.coroutine
        def coro():
            pass

        self.assertIs(
            coro,
            self.decorator(coro),
        )

    def test_adds_magic_attribute(self):
        @asyncio.coroutine
        def coro():
            pass

        self.decorator(coro)

        self.assertTrue(hasattr(coro, "_aioxmpp_service_handlers"))

    def test_adds__apply_iq_handler_entry(self):
        @asyncio.coroutine
        def coro():
            pass

        self.decorator(coro)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_iq_handler,
                 (unittest.mock.sentinel.type_,
                  self.Payload)),
                is_unique=True,
                require_deps=(),
            ),
            coro._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        @asyncio.coroutine
        def coro():
            pass

        coro._aioxmpp_service_handlers = {"foo"}

        self.decorator(coro)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_iq_handler,
                 (unittest.mock.sentinel.type_,
                  self.Payload)),
                is_unique=True,
                require_deps=(),
            ),
            coro._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            coro._aioxmpp_service_handlers,
        )

    def test_accepts_normal_function(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = False

            self.decorator(unittest.mock.sentinel.coro)

        iscoroutinefunction.assert_not_called()

    def test_works_with_is_iq_handler(self):
        @asyncio.coroutine
        def coro():
            pass

        self.assertFalse(
            service.is_iq_handler(
                unittest.mock.sentinel.type_,
                self.Payload,
                coro,
            )
        )

        self.decorator(coro)

        self.assertTrue(
            service.is_iq_handler(
                unittest.mock.sentinel.type_,
                self.Payload,
                coro,
            )
        )


class Testmessage_handler(unittest.TestCase):
    def test_forwards_to_dispatcher(self):
        with unittest.mock.patch(
                "aioxmpp.dispatcher.message_handler") as message_handler:
            result = service.message_handler(
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
            )

        message_handler.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
        )

        self.assertEqual(
            result,
            message_handler(),
        )


class Testpresence_handler(unittest.TestCase):
    def test_forwards_to_dispatcher(self):
        with unittest.mock.patch(
                "aioxmpp.dispatcher.presence_handler") as presence_handler:
            result = service.presence_handler(
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
            )

        presence_handler.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
        )

        self.assertEqual(
            result,
            presence_handler(),
        )


class Testinbound_message_filter(unittest.TestCase):
    def setUp(self):
        self.decorator = service.inbound_message_filter

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_inbound_message_filter_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_inbound_message_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_inbound_message_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )


class Testinbound_presence_filter(unittest.TestCase):
    def setUp(self):
        self.decorator = service.inbound_presence_filter

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_inbound_presence_filter_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_inbound_presence_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_inbound_presence_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )


class Testoutbound_message_filter(unittest.TestCase):
    def setUp(self):
        self.decorator = service.outbound_message_filter

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_outbound_message_filter_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_outbound_message_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_outbound_message_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )


class Testoutbound_presence_filter(unittest.TestCase):
    def setUp(self):
        self.decorator = service.outbound_presence_filter

    def tearDown(self):
        del self.decorator

    def test_works_as_decorator(self):
        def cb():
            pass

        self.assertIs(
            cb,
            self.decorator(cb),
        )

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_adds__apply_outbound_presence_filter_entry(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_outbound_presence_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (service._apply_outbound_presence_filter, ()),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_requires_non_coroutine(self):
        with unittest.mock.patch(
                "asyncio.iscoroutinefunction") as iscoroutinefunction:
            iscoroutinefunction.return_value = True

            with self.assertRaisesRegex(
                    TypeError,
                    "must not be a coroutine function"):
                self.decorator(unittest.mock.sentinel.cb)

        iscoroutinefunction.assert_called_with(
            unittest.mock.sentinel.cb,
        )


class Testdepsignal(unittest.TestCase):
    class S1(service.Service):
        signal = callbacks.Signal()
        sync = callbacks.SyncSignal()

    def setUp(self):
        self.decorator = service.depsignal(
            self.S1,
            "signal",
        )

    def tearDown(self):
        del self.decorator

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_uses_strong_by_default(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depsignal,
                    (
                        self.S1,
                        "signal",
                        callbacks.AdHocSignal.STRONG,
                    )
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_defer_flag(self):
        self.decorator = service.depsignal(
            self.S1,
            "signal",
            defer=True,
        )

        def cb():
            pass

        with unittest.mock.patch.object(
                callbacks.AdHocSignal,
                "ASYNC_WITH_LOOP") as ASYNC_WITH_LOOP:
            ASYNC_WITH_LOOP.return_value = \
                unittest.mock.sentinel.async_with_loop
            self.decorator(cb)

        ASYNC_WITH_LOOP.assert_not_called()

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depsignal,
                    (
                        self.S1,
                        "signal",
                        (ASYNC_WITH_LOOP, (None,)),
                    )
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_require_coroutinefunction_for_sync_signal(self):
        def cb():
            pass

        self.decorator = service.depsignal(
            self.S1,
            "sync",
        )

        with self.assertRaisesRegex(
                TypeError,
                "a coroutine function is required for this signal"):
            self.decorator(cb)

    def test_coroutinefunction_and_sync_signal(self):
        @asyncio.coroutine
        def coro():
            pass

        self.decorator = service.depsignal(
            self.S1,
            "sync",
        )

        self.decorator(coro)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depsignal,
                    (
                        self.S1,
                        "sync",
                        None,
                    )
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            coro._aioxmpp_service_handlers
        )

    def test_use_spawn_for_coroutinefunction_on_normal_signal_and_defer(self):
        self.decorator = service.depsignal(
            self.S1,
            "signal",
            defer=True,
        )

        @asyncio.coroutine
        def coro():
            pass

        with unittest.mock.patch.object(
                callbacks.AdHocSignal,
                "SPAWN_WITH_LOOP") as SPAWN_WITH_LOOP:
            SPAWN_WITH_LOOP.return_value = \
                unittest.mock.sentinel.spawn_with_loop
            self.decorator(coro)

        SPAWN_WITH_LOOP.assert_not_called()

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depsignal,
                    (
                        self.S1,
                        "signal",
                        (SPAWN_WITH_LOOP, (None,))
                    )
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            coro._aioxmpp_service_handlers
        )

    def test_reject_coroutinefunction_on_normal_signal_without_defer(self):
        self.decorator = service.depsignal(
            self.S1,
            "signal",
        )

        @asyncio.coroutine
        def coro():
            pass

        with self.assertRaisesRegex(
                TypeError,
                "cannot use coroutine function with this signal "
                "without defer"):
            self.decorator(coro)

    def test_reject_defer_on_sync_signal(self):
        self.decorator = service.depsignal(
            self.S1,
            "sync",
            defer=True,
        )

        @asyncio.coroutine
        def coro():
            pass

        with self.assertRaisesRegex(
                ValueError,
                "cannot use defer with this signal"):
            self.decorator(coro)

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depsignal,
                    (
                        self.S1,
                        "signal",
                        callbacks.AdHocSignal.STRONG,
                    ),
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_adds_dependency(self):
        def cb():
            pass

        self.decorator(cb)

        spec, = cb._aioxmpp_service_handlers
        self.assertIn(
            self.S1,
            spec.require_deps,
        )

    def test_does_not_add_dependency_for_StanzaStream(self):
        def cb():
            pass

        decorator = service.depsignal(
            aioxmpp.stream.StanzaStream,
            "on_message_received",
        )
        decorator(cb)

        spec, = cb._aioxmpp_service_handlers
        self.assertNotIn(
            aioxmpp.stream.StanzaStream,
            spec.require_deps,
        )

    def test_does_not_add_dependency_for_Client(self):
        @asyncio.coroutine
        def cb():
            pass

        decorator = service.depsignal(
            aioxmpp.node.Client,
            "before_stream_established",
        )
        decorator(cb)

        spec, = cb._aioxmpp_service_handlers
        self.assertNotIn(
            aioxmpp.node.Client,
            spec.require_deps,
        )


class Testdepfilter(unittest.TestCase):
    class S1:
        pass

    def setUp(self):
        self.decorator = service.depfilter(
            self.S1,
            "filter",
        )

    def tearDown(self):
        del self.decorator

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_depfilter,
                    (
                        self.S1,
                        "filter",
                    ),
                ),
                is_unique=True,
                require_deps=(self.S1,),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )

    def test_adds_dependency(self):
        def cb():
            pass

        self.decorator(cb)

        spec, = cb._aioxmpp_service_handlers
        self.assertIn(
            self.S1,
            spec.require_deps,
        )

    def test_does_not_add_dependency_for_StanzaStream(self):
        def cb():
            pass

        decorator = service.depfilter(
            aioxmpp.stream.StanzaStream,
            "some_filter",
        )
        decorator(cb)

        spec, = cb._aioxmpp_service_handlers
        self.assertNotIn(
            aioxmpp.stream.StanzaStream,
            spec.require_deps,
        )


class Testattrsignal(unittest.TestCase):
    class DescriptorValue:
        signal = callbacks.Signal()
        sync = callbacks.SyncSignal()


    class Descriptor(service.Descriptor):
        def init_cm(self, instance):
            raise NotImplementedError

        @property
        def value_type(self):
            return Testattrsignal.DescriptorValue

    def setUp(self):
        self.descriptor = self.Descriptor()

        self.decorator = service.attrsignal(
            self.descriptor,
            "signal",
        )

    def tearDown(self):
        del self.decorator

    def test_adds_magic_attribute(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertTrue(hasattr(cb, "_aioxmpp_service_handlers"))

    def test_uses_strong_by_default(self):
        def cb():
            pass

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_attrsignal,
                    (
                        self.descriptor,
                        "signal",
                        callbacks.AdHocSignal.STRONG,
                    )
                ),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_defer_flag(self):
        self.decorator = service.attrsignal(
            self.descriptor,
            "signal",
            defer=True,
        )

        def cb():
            pass

        with unittest.mock.patch.object(
                callbacks.AdHocSignal,
                "ASYNC_WITH_LOOP") as ASYNC_WITH_LOOP:
            ASYNC_WITH_LOOP.return_value = \
                unittest.mock.sentinel.async_with_loop
            self.decorator(cb)

        ASYNC_WITH_LOOP.assert_not_called()

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_attrsignal,
                    (
                        self.descriptor,
                        "signal",
                        (ASYNC_WITH_LOOP, (None,))
                    )
                ),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

    def test_require_coroutinefunction_for_sync_signal(self):
        def cb():
            pass

        self.decorator = service.attrsignal(
            self.descriptor,
            "sync",
        )

        with self.assertRaisesRegex(
                TypeError,
                "a coroutine function is required for this signal"):
            self.decorator(cb)

    def test_coroutinefunction_and_sync_signal(self):
        @asyncio.coroutine
        def coro():
            pass

        self.decorator = service.attrsignal(
            self.descriptor,
            "sync",
        )

        self.decorator(coro)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_attrsignal,
                    (
                        self.descriptor,
                        "sync",
                        None,
                    )
                ),
                is_unique=True,
                require_deps=(),
            ),
            coro._aioxmpp_service_handlers
        )

    def test_use_spawn_for_coroutinefunction_on_normal_signal_and_defer(self):
        self.decorator = service.attrsignal(
            self.descriptor,
            "signal",
            defer=True,
        )

        @asyncio.coroutine
        def coro():
            pass

        with unittest.mock.patch.object(
                callbacks.AdHocSignal,
                "SPAWN_WITH_LOOP") as SPAWN_WITH_LOOP:
            SPAWN_WITH_LOOP.return_value = \
                unittest.mock.sentinel.spawn_with_loop
            self.decorator(coro)

        SPAWN_WITH_LOOP.assert_not_called()

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_attrsignal,
                    (
                        self.descriptor,
                        "signal",
                        (SPAWN_WITH_LOOP, (None,))
                    )
                ),
                is_unique=True,
                require_deps=(),
            ),
            coro._aioxmpp_service_handlers
        )

    def test_reject_coroutinefunction_on_normal_signal_without_defer(self):
        self.decorator = service.attrsignal(
            self.descriptor,
            "signal",
        )

        @asyncio.coroutine
        def coro():
            pass

        with self.assertRaisesRegex(
                TypeError,
                "cannot use coroutine function with this signal "
                "without defer"):
            self.decorator(coro)

    def test_reject_defer_on_sync_signal(self):
        self.decorator = service.attrsignal(
            self.descriptor,
            "sync",
            defer=True,
        )

        @asyncio.coroutine
        def coro():
            pass

        with self.assertRaisesRegex(
                ValueError,
                "cannot use defer with this signal"):
            self.decorator(coro)

    def test_stacks_with_other_effects(self):
        def cb():
            pass

        cb._aioxmpp_service_handlers = {"foo"}

        self.decorator(cb)

        self.assertIn(
            service.HandlerSpec(
                (
                    service._apply_connect_attrsignal,
                    (
                        self.descriptor,
                        "signal",
                        callbacks.AdHocSignal.STRONG,
                    ),
                ),
                is_unique=True,
                require_deps=(),
            ),
            cb._aioxmpp_service_handlers
        )

        self.assertIn(
            "foo",
            cb._aioxmpp_service_handlers,
        )


class Testis_iq_handler(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_iq_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.payload_cls,
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_iq_handler,
                 (unittest.mock.sentinel.type_,
                  unittest.mock.sentinel.payload_cls)),
            )
        ]

        self.assertTrue(
            service.is_iq_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.payload_cls,
                m
            )
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_iq_handler,
                 (unittest.mock.sentinel.type2,
                  unittest.mock.sentinel.payload_cls2)),
            )
        ]

        self.assertFalse(
            service.is_iq_handler(
                unittest.mock.sentinel.type_,
                unittest.mock.sentinel.payload_cls,
                m
            )
        )


class Testis_message_handler(unittest.TestCase):
    def test_forwards_to_dispatcher(self):
        with unittest.mock.patch(
                "aioxmpp.dispatcher."
                "is_message_handler") as is_message_handler:
            result = service.is_message_handler(
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
                unittest.mock.sentinel.c,
            )

        is_message_handler.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.c,
        )

        self.assertEqual(
            result,
            is_message_handler(),
        )


class Testis_presence_handler(unittest.TestCase):
    def test_forwards_to_dispatcher(self):
        with unittest.mock.patch(
                "aioxmpp.dispatcher."
                "is_presence_handler") as is_presence_handler:
            result = service.is_presence_handler(
                unittest.mock.sentinel.a1,
                unittest.mock.sentinel.a2,
                unittest.mock.sentinel.c,
            )

        is_presence_handler.assert_called_once_with(
            unittest.mock.sentinel.a1,
            unittest.mock.sentinel.a2,
            unittest.mock.sentinel.c,
        )

        self.assertEqual(
            result,
            is_presence_handler(),
        )


class Testis_inbound_message_filter(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_inbound_message_filter(
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_inbound_message_filter,
                 ())
            )
        ]

        self.assertTrue(
            service.is_inbound_message_filter(m)
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_inbound_presence_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_inbound_message_filter(m)
        )


class Testis_inbound_presence_filter(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_inbound_presence_filter(
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_inbound_presence_filter,
                 ())
            )
        ]

        self.assertTrue(
            service.is_inbound_presence_filter(m)
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_inbound_message_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_inbound_presence_filter(m)
        )


class Testis_outbound_message_filter(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_outbound_message_filter(
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_message_filter,
                 ())
            )
        ]

        self.assertTrue(
            service.is_outbound_message_filter(m)
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_presence_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_outbound_message_filter(m)
        )


class Testis_outbound_presence_filter(unittest.TestCase):
    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_outbound_presence_filter(
                object()
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_presence_filter,
                 ())
            )
        ]

        self.assertTrue(
            service.is_outbound_presence_filter(m)
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_message_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_outbound_presence_filter(m)
        )


class Testis_depsignal_handler(unittest.TestCase):
    class S1(service.Service):
        signal = callbacks.Signal()
        sync = callbacks.SyncSignal()

    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_depsignal_handler(
                self.S1,
                "signal",
                object(),
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        def cb(self):
            pass

        @asyncio.coroutine
        def coro(self):
            pass

        tokens = [
            (
                "signal",
                False,
                cb,
                service.HandlerSpec(
                    (
                        service._apply_connect_depsignal,
                        (
                            self.S1,
                            "signal",
                            callbacks.AdHocSignal.STRONG,
                        ),
                    ),
                    require_deps=(self.S1,)
                )
            ),
            (
                "signal",
                True,
                cb,
                service.HandlerSpec(
                    (
                        service._apply_connect_depsignal,
                        (
                            self.S1,
                            "signal",
                            (unittest.mock.sentinel.async_with_loop, (None,))
                        ),
                    ),
                    require_deps=(self.S1,)
                )
            ),
            (
                "signal",
                True,
                coro,
                service.HandlerSpec(
                    (
                        service._apply_connect_depsignal,
                        (
                            self.S1,
                            "signal",
                            (unittest.mock.sentinel.spawn_with_loop, (None,))
                        ),
                    ),
                    require_deps=(self.S1,)
                )
            ),
            (
                "sync",
                False,
                coro,
                service.HandlerSpec(
                    (
                        service._apply_connect_depsignal,
                        (
                            self.S1,
                            "sync",
                            None,
                        ),
                    ),
                    require_deps=(self.S1,),
                )
            ),
        ]

        for signal_name, defer, obj, spec in tokens:
            with contextlib.ExitStack() as stack:
                SPAWN_WITH_LOOP = stack.enter_context(
                    unittest.mock.patch.object(
                        callbacks.AdHocSignal,
                        "SPAWN_WITH_LOOP",
                        new=unittest.mock.sentinel.spawn_with_loop)
                )


                ASYNC_WITH_LOOP = stack.enter_context(
                    unittest.mock.patch.object(
                        callbacks.AdHocSignal,
                        "ASYNC_WITH_LOOP",
                        new=unittest.mock.sentinel.async_with_loop)
                )


                obj._aioxmpp_service_handlers = [
                    spec,
                ]

                self.assertTrue(
                    service.is_depsignal_handler(
                        self.S1,
                        signal_name,
                        obj,
                        defer=defer,
                    ),
                    spec,
                )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_message_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_depsignal_handler(
                self.S1,
                "signal",
                m,
                defer=True,
            )
        )

    def test_works_for_deferred_coroutinefunctions(self):
        class Cls:
            signal = aioxmpp.callbacks.Signal()

        @service.depsignal(Cls, "signal", defer=True)
        @asyncio.coroutine
        def coro():
            pass

        self.assertTrue(
            service.is_depsignal_handler(
                Cls,
                "signal",
                coro,
                defer=True,
            )
        )

    def test_works_for_deferred_functions(self):
        class Cls:
            signal = aioxmpp.callbacks.Signal()

        @service.depsignal(Cls, "signal", defer=True)
        def func():
            pass

        self.assertTrue(
            service.is_depsignal_handler(
                Cls,
                "signal",
                func,
                defer=True,
            )
        )


class Testis_depfilter_handler(unittest.TestCase):
    class S1:
        pass

    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_depfilter_handler(
                self.S1,
                "filter",
                object(),
            )
        )

    def test_works_with_depfilter(self):
        def f():
            pass

        service.depfilter(self.S1, "filter")(f)

        self.assertTrue(
            service.is_depfilter_handler(
                self.S1,
                "filter",
                f,
            )
        )

        self.assertFalse(
            service.is_depfilter_handler(
                self.S1,
                "foo",
                f,
            )
        )

        self.assertFalse(
            service.is_depfilter_handler(
                aioxmpp.stream.StanzaStream,
                "foo",
                f,
            )
        )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_message_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_depfilter_handler(
                self.S1,
                "filter",
                m,
            )
        )


class Testis_attrsignal_handler(unittest.TestCase):
    class DescriptorValue:
        signal = callbacks.Signal()
        sync = callbacks.SyncSignal()


    class Descriptor(service.Descriptor):
        def init_cm(self, instance):
            raise NotImplementedError

        @property
        def value_type(self):
            return Testattrsignal.DescriptorValue

    def setUp(self):
        self.descriptor = self.Descriptor()

    def test_return_false_if_magic_attr_is_missing(self):
        self.assertFalse(
            service.is_attrsignal_handler(
                self.descriptor,
                "signal",
                object(),
            )
        )

    def test_return_true_if_token_in_magic_attr(self):
        def cb(self):
            pass

        @asyncio.coroutine
        def coro(self):
            pass

        tokens = [
            (
                "signal",
                False,
                cb,
                service.HandlerSpec(
                    (
                        service._apply_connect_attrsignal,
                        (
                            self.descriptor,
                            "signal",
                            callbacks.AdHocSignal.STRONG,
                        ),
                    ),
                    require_deps=()
                )
            ),
            (
                "signal",
                True,
                cb,
                service.HandlerSpec(
                    (
                        service._apply_connect_attrsignal,
                        (
                            self.descriptor,
                            "signal",
                            (unittest.mock.sentinel.async_with_loop, (None,))
                        ),
                    ),
                    require_deps=()
                )
            ),
            (
                "signal",
                True,
                coro,
                service.HandlerSpec(
                    (
                        service._apply_connect_attrsignal,
                        (
                            self.descriptor,
                            "signal",
                            (unittest.mock.sentinel.spawn_with_loop, (None,))
                        ),
                    ),
                    require_deps=()
                )
            ),
            (
                "sync",
                False,
                coro,
                service.HandlerSpec(
                    (
                        service._apply_connect_attrsignal,
                        (
                            self.descriptor,
                            "sync",
                            None,
                        ),
                    ),
                    require_deps=(),
                )
            ),
        ]

        for signal_name, defer, obj, spec in tokens:
            with contextlib.ExitStack() as stack:
                SPAWN_WITH_LOOP = stack.enter_context(
                    unittest.mock.patch.object(
                        callbacks.AdHocSignal,
                        "SPAWN_WITH_LOOP",
                        new=unittest.mock.sentinel.spawn_with_loop)
                )

                ASYNC_WITH_LOOP = stack.enter_context(
                    unittest.mock.patch.object(
                        callbacks.AdHocSignal,
                        "ASYNC_WITH_LOOP",
                        new=unittest.mock.sentinel.async_with_loop)
                )

                obj._aioxmpp_service_handlers = [
                    spec,
                ]

                self.assertTrue(
                    service.is_attrsignal_handler(
                        self.descriptor,
                        signal_name,
                        obj,
                        defer=defer,
                    ),
                    spec,
                )

    def test_return_false_if_token_not_in_magic_attr(self):
        m = unittest.mock.Mock()
        m._aioxmpp_service_handlers = [
            service.HandlerSpec(
                (service._apply_outbound_message_filter,
                 ())
            )
        ]

        self.assertFalse(
            service.is_attrsignal_handler(
                self.descriptor,
                "signal",
                m,
                defer=True,
            )
        )

    def test_works_for_deferred_coroutinefunctions(self):
        @service.attrsignal(self.descriptor, "signal", defer=True)
        @asyncio.coroutine
        def coro():
            pass

        self.assertTrue(
            service.is_attrsignal_handler(
                self.descriptor,
                "signal",
                coro,
                defer=True,
            )
        )

    def test_works_for_deferred_functions(self):
        @service.attrsignal(self.descriptor, "signal", defer=True)
        def func():
            pass

        self.assertTrue(
            service.is_attrsignal_handler(
                self.descriptor,
                "signal",
                func,
                defer=True,
            )
        )


class TestDescriptor(unittest.TestCase):
    class DescriptorSubclass(service.Descriptor):
        def init_cm(self, instance):
            return unittest.mock.sentinel.cm

        @property
        def value_type(self):
            return None

    def setUp(self):
        self.d = self.DescriptorSubclass()

    def tearDown(self):
        del self.d

    def test_is_abstract_class(self):
        with self.assertRaisesRegex(TypeError, "abstract methods init_cm"):
            service.Descriptor()

    def test_add_to_stack_uses_init_cm_to_obtain_cm_and_pushes(self):
        with contextlib.ExitStack() as stack:
            init_cm = stack.enter_context(
                unittest.mock.patch.object(self.d, "init_cm")
            )
            init_cm.return_value = unittest.mock.sentinel.cm

            target_stack = unittest.mock.Mock()

            result = self.d.add_to_stack(
                unittest.mock.sentinel.instance,
                target_stack,
            )

        init_cm.assert_called_once_with(
            unittest.mock.sentinel.instance,
        )

        target_stack.enter_context.assert_called_once_with(
            unittest.mock.sentinel.cm,
        )

        self.assertEqual(
            result,
            target_stack.enter_context(),
        )

    def test___get___returns_self_for_None_instance(self):
        self.assertIs(
            self.d.__get__(None, unittest.mock.sentinel.owner),
            self.d,
        )

    def test___get___raises_AttributeError_if_not_initialised(self):
        with self.assertRaisesRegex(
                AttributeError,
                r"resource manager descriptor has not been initialised"):
            self.d.__get__(
                unittest.mock.sentinel.instance,
                unittest.mock.sentinel.owner
            )

    def test_add_to_stack_causes___get___to_return_the_cm_result(self):
        target_stack = unittest.mock.Mock()

        result1 = self.d.add_to_stack(
            unittest.mock.sentinel.instance,
            target_stack,
        )

        result2 = self.d.__get__(
            unittest.mock.sentinel.instance,
            unittest.mock.sentinel.owner,
        )

        self.assertEqual(result1, result2)

    def test___get___works_with_different_instances(self):
        target_stack = unittest.mock.Mock()

        result11 = self.d.add_to_stack(
            unittest.mock.sentinel.instance1,
            target_stack,
        )

        result12 = self.d.add_to_stack(
            unittest.mock.sentinel.instance2,
            target_stack,
        )

        result21 = self.d.__get__(
            unittest.mock.sentinel.instance1,
            unittest.mock.sentinel.owner,
        )

        result22 = self.d.__get__(
            unittest.mock.sentinel.instance2,
            unittest.mock.sentinel.owner,
        )

        with self.assertRaises(AttributeError):
            self.d.__get__(
                unittest.mock.sentinel.instance3,
                unittest.mock.sentinel.owner,
            )

        self.assertEqual(result11, result21)
        self.assertEqual(result12, result22)

    def test_add_to_stack_stores_with_weakref(self):
        with contextlib.ExitStack() as stack:
            WeakKeyDictionary = stack.enter_context(
                unittest.mock.patch(
                    "weakref.WeakKeyDictionary",
                    new=unittest.mock.MagicMock(),
                ),
            )

            target_stack = unittest.mock.Mock()

            self.d = self.DescriptorSubclass()

            WeakKeyDictionary.assert_called_once_with()

            self.d.add_to_stack(
                unittest.mock.sentinel.instance,
                target_stack,
            )

            self.assertIn(
                unittest.mock._Call((
                    "__setitem__",
                    (
                        unittest.mock.sentinel.instance,
                        (
                            unittest.mock.sentinel.cm,
                            target_stack.enter_context(),
                        )
                    ),
                    {}
                )),
                WeakKeyDictionary().mock_calls,
            )



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.roster.xso as roster_xso
import aioxmpp.stanza as stanza
import aioxmpp.nonza as nonza
import aioxmpp.structs as structs
import aioxmpp.xso as xso
import aioxmpp.xso.model as xso_model

from aioxmpp.utils import namespaces


TEST_JID = structs.JID.fromstr("foo@bar.example")


class TestNamespaces(unittest.TestCase):
    def test_roster_namespace(self):
        self.assertEqual(
            "jabber:iq:roster",
            namespaces.rfc6121_roster
        )

    def test_roster_versioning_namespace(self):
        self.assertEqual(
            "urn:xmpp:features:rosterver",
            namespaces.rfc6121_roster_versioning
        )


class TestGroup(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.rfc6121_roster, "group"),
            roster_xso.Group.TAG
        )

    def test_name_attr(self):
        self.assertIsInstance(
            roster_xso.Group.name,
            xso.Text
        )

    def test_init(self):
        g = roster_xso.Group()
        self.assertIsNone(g.name)

        g = roster_xso.Group(name="foobar")
        self.assertEqual("foobar", g.name)


class TestItem(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.rfc6121_roster, "item"),
            roster_xso.Item.TAG
        )

    def test_approved_attr(self):
        self.assertIsInstance(
            roster_xso.Item.approved,
            xso.Attr
        )
        self.assertEqual(
            (None, "approved"),
            roster_xso.Item.approved.tag
        )
        self.assertIsInstance(
            roster_xso.Item.approved.type_,
            xso.Bool
        )
        self.assertEqual(
            False,
            roster_xso.Item.approved.default
        )

    def test_ask_attr(self):
        self.assertIsInstance(
            roster_xso.Item.ask,
            xso.Attr
        )
        self.assertEqual(
            (None, "ask"),
            roster_xso.Item.ask.tag
        )
        self.assertIsInstance(
            roster_xso.Item.ask.validator,
            xso.RestrictToSet
        )
        self.assertEqual(
            xso.ValidateMode.ALWAYS,
            roster_xso.Item.ask.validate
        )
        self.assertSetEqual(
            {
                None,
                "subscribe",
            },
            roster_xso.Item.ask.validator.values
        )
        self.assertEqual(
            None,
            roster_xso.Item.ask.default
        )

    def test_jid_attr(self):
        self.assertIsInstance(
            roster_xso.Item.jid,
            xso.Attr
        )
        self.assertEqual(
            (None, "jid"),
            roster_xso.Item.jid.tag
        )
        self.assertIsInstance(
            roster_xso.Item.jid.type_,
            xso.JID
        )

    def test_name_attr(self):
        self.assertIsInstance(
            roster_xso.Item.name,
            xso.Attr
        )
        self.assertEqual(
            (None, "name"),
            roster_xso.Item.name.tag
        )

    def test_subscription_attr(self):
        self.assertIsInstance(
            roster_xso.Item.subscription,
            xso.Attr
        )
        self.assertEqual(
            (None, "subscription"),
            roster_xso.Item.subscription.tag
        )
        self.assertIsInstance(
            roster_xso.Item.subscription.validator,
            xso.RestrictToSet
        )
        self.assertEqual(
            xso.ValidateMode.ALWAYS,
            roster_xso.Item.subscription.validate
        )
        self.assertEqual(
            roster_xso.Item.subscription.default,
            "none",
        )
        self.assertSetEqual(
            {
                "none",
                "to",
                "from",
                "both",
                "remove",
            },
            roster_xso.Item.subscription.validator.values
        )

    def test_groups_attr(self):
        self.assertIsInstance(
            roster_xso.Item.groups,
            xso.ChildList
        )
        self.assertSetEqual(
            {
                roster_xso.Group
            },
            set(roster_xso.Item.groups._classes)
        )

    def test_init(self):
        jid = structs.JID.fromstr("foo@bar.example")
        item = roster_xso.Item(jid)
        self.assertEqual(item.jid, jid)
        self.assertIsNone(item.name)
        self.assertSequenceEqual([], item.groups)
        self.assertEqual("none", item.subscription)
        self.assertIs(False, item.approved)
        self.assertIsNone(item.ask)

        group = roster_xso.Group()
        item = roster_xso.Item(
            jid=jid,
            name="foobar",
            groups=(group,),
            subscription="to",
            approved=True,
            ask="subscribe"
        )

        self.assertEqual(jid, item.jid)
        self.assertEqual("foobar", item.name)
        self.assertSequenceEqual([group], item.groups)
        self.assertIsInstance(item.groups, xso_model.XSOList)
        self.assertEqual("to", item.subscription)
        self.assertIs(True, item.approved)
        self.assertEqual("subscribe", item.ask)


class TestQuery(unittest.TestCase):
    def test_tag(self):
        self.assertEqual(
            (namespaces.rfc6121_roster, "query"),
            roster_xso.Query.TAG
        )

    def test_declare_ns(self):
        self.assertDictEqual(
            roster_xso.Query.DECLARE_NS,
            {
                None: namespaces.rfc6121_roster
            }
        )

    def test_ver_attr(self):
        self.assertIsInstance(
            roster_xso.Query.ver,
            xso.Attr
        )
        self.assertEqual(
            (None, "ver"),
            roster_xso.Query.ver.tag
        )
        self.assertIs(
            roster_xso.Query.ver.default,
            None
        )

    def test_items_attr(self):
        self.assertIsInstance(
            roster_xso.Query.items,
            xso.ChildList
        )
        self.assertSetEqual(
            {
                roster_xso.Item
            },
            set(roster_xso.Query.items._classes)
        )

    def test_init(self):
        q = roster_xso.Query()
        self.assertIsNone(q.ver)
        self.assertSequenceEqual([], q.items)

        item = roster_xso.Item(TEST_JID)
        q = roster_xso.Query(
            ver="foobar",
            items=(item,)
        )
        self.assertEqual("foobar", q.ver)
        self.assertSequenceEqual([item], q.items)
        self.assertIsInstance(q.items, xso_model.XSOList)

    def test_is_registered_as_iq_payload(self):
        self.assertIn(
            roster_xso.Query.TAG,
            stanza.IQ.CHILD_MAP
        )


class TestRosterVersioningFeature(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            roster_xso.RosterVersioningFeature,
            xso.XSO))

    def test_tag(self):
        self.assertEqual(
            (namespaces.rfc6121_roster_versioning, "ver"),
            roster_xso.RosterVersioningFeature.TAG
        )

    def test_is_registered_as_feature(self):
        self.assertIn(
            roster_xso.RosterVersioningFeature.TAG,
            nonza.StreamFeatures.CHILD_MAP
        )



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.roster` --- :rfc:`6121` roster implementation
############################################################

This subpackage provides :class:`.RosterClient`, a service to interact with
:rfc:`6121` rosters.

.. currentmodule:: aioxmpp

.. autoclass:: RosterClient

.. currentmodule:: aioxmpp.roster

.. class:: Service

   Alias of :class:`.RosterClient`.

   .. deprecated:: 0.8

      The alias will be removed in 1.0.

.. autoclass:: Item

.. module:: aioxmpp.roster.xso

.. currentmodule:: aioxmpp.roster.xso

:mod:`.roster.xso` --- IQ payloads and stream feature
=====================================================

The submodule :mod:`aioxmpp.roster.xso` contains the :class:`~aioxmpp.xso.XSO`
classes which describe the IQ payloads used by this subpackage.

.. autoclass:: Query

.. autoclass:: Item

.. autoclass:: Group

The stream feature which is used by servers to announce support for roster
versioning:

.. autoclass:: RosterVersioningFeature()
"""

from .service import RosterClient, Item  # NOQA
Service = RosterClient  # NOQA



########################################################################
# File name: test_fields.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import abc
import collections
import contextlib
import copy
import itertools
import unittest
import unittest.mock

import aioxmpp
import aioxmpp.xso as xso

import aioxmpp.forms.fields as fields
import aioxmpp.forms.xso as forms_xso


TEST_JID = aioxmpp.JID.fromstr("foo@bar.baz/fnord")


def instance_mock():
    mock = unittest.mock.Mock()
    mock._descriptor_data = {}
    return mock


def generate_values(prefix):
    i = 0
    while True:
        yield getattr(unittest.mock.sentinel, prefix + str(i))
        i += 1


class TestBoundField(unittest.TestCase):
    class FakeBoundField(fields.BoundField):
        def load(self, field_xso):
            return super().load(field_xso)

        def render(self, **kwargs):
            return super().render(**kwargs)

    def setUp(self):
        self.field = unittest.mock.Mock()
        self.instance = unittest.mock.sentinel.instance
        self.bf = self.FakeBoundField(
            self.field,
            self.instance
        )

    def tearDown(self):
        del self.instance

    def test_is_abstract(self):
        with self.assertRaisesRegex(TypeError, r"abstract.* load, render"):
            fields.BoundField(unittest.mock.sentinel.foo,
                              unittest.mock.sentinel.bar)

    def test_repr(self):
        self.assertRegex(
            repr(self.bf),
            r"<bound <Mock .*> for sentinel\.instance at 0x[0-9a-f]+>"
        )

    def test_reading_desc_reads_from_field(self):
        self.assertEqual(
            self.bf.desc,
            self.field.desc,
        )

    def test_setting_desc_stores_locally(self):
        self.bf.desc = unittest.mock.sentinel.desc

        self.assertEqual(
            self.bf.desc,
            unittest.mock.sentinel.desc,
        )

        self.assertNotEqual(
            self.field.desc,
            unittest.mock.sentinel.desc,
        )

    def test_reading_label_reads_from_field(self):
        self.assertEqual(
            self.bf.label,
            self.field.label,
        )

    def test_setting_label_stores_locally(self):
        self.bf.label = unittest.mock.sentinel.label

        self.assertEqual(
            self.bf.label,
            unittest.mock.sentinel.label,
        )

        self.assertNotEqual(
            self.field.label,
            unittest.mock.sentinel.label,
        )

    def test_reading_required_reads_from_field(self):
        self.assertEqual(
            self.bf.required,
            self.field.required,
        )

    def test_setting_required_stores_locally(self):
        self.bf.required = unittest.mock.sentinel.required

        self.assertEqual(
            self.bf.required,
            unittest.mock.sentinel.required,
        )

        self.assertNotEqual(
            self.field.required,
            unittest.mock.sentinel.required,
        )

    def test_instance_is_settable(self):
        self.assertEqual(
            self.bf.instance,
            self.instance,
        )

        self.bf.instance = unittest.mock.sentinel.other_instance

        self.assertRegex(
            repr(self.bf),
            r"<bound <Mock .*> for sentinel\.other_instance at 0x[0-9a-f]+>"
        )

    def test_field_is_not(self):
        self.assertEqual(
            self.bf.field,
            self.field,
        )

        with self.assertRaises(AttributeError):
            self.bf.field = "foo"

    def test_deepcopy_leaves_instance_and_field_identical(self):
        self.bf.desc = ["foo"]

        copied = copy.deepcopy(self.bf)

        self.assertIsNot(copied.desc, self.bf.desc)

        self.assertIs(copied.instance, self.bf.instance)
        self.assertIs(copied.field, self.bf.field)

    def test_clone_for(self):
        with unittest.mock.patch("copy.deepcopy") as deepcopy:
            other_bf = self.bf.clone_for(
                unittest.mock.sentinel.other_instance
            )

        deepcopy.assert_called_with(self.bf)

        self.assertEqual(
            other_bf,
            deepcopy(),
        )

        self.assertEqual(
            other_bf.instance,
            unittest.mock.sentinel.other_instance,
        )

    def test_clone_for_passes_memo(self):
        with unittest.mock.patch("copy.deepcopy") as deepcopy:
            other_bf = self.bf.clone_for(
                unittest.mock.sentinel.other_instance,
                memo=unittest.mock.sentinel.memo,
            )

        deepcopy.assert_called_with(
            self.bf,
            unittest.mock.sentinel.memo
        )

        self.assertEqual(
            other_bf,
            deepcopy(),
        )

        self.assertEqual(
            other_bf.instance,
            unittest.mock.sentinel.other_instance,
        )

    def test_load_loads_desc(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            desc="foobar",
        )

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.desc,
            "foobar",
        )

    def test_load_leaves_desc_unmodified_if_unset(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
        )

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.desc,
            self.field.desc
        )

        self.bf.desc = "foobar"

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.desc,
            "foobar",
        )

    def test_load_loads_label(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            label="foobar"
        )

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.label,
            "foobar",
        )

    def test_load_leaves_label_unmodified_if_unset(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
        )

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.label,
            self.field.label
        )

        self.bf.label = "foobar"

        self.bf.load(field_xso)

        self.assertEqual(
            self.bf.label,
            "foobar",
        )

    def test_load_overrides_required(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            required=True,
        )

        self.bf.load(field_xso)

        self.assertIs(
            self.bf.required,
            True,
        )

        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            required=False,
        )

        self.bf.load(field_xso)

        self.assertIs(
            self.bf.required,
            False,
        )

    def test_render_defers_to_field_by_default(self):
        self.field.desc = "some description"
        self.field.label = "some label"
        self.field.required = unittest.mock.sentinel.required

        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"

        field_xso = self.bf.render()

        self.assertIsInstance(
            field_xso,
            forms_xso.Field,
        )

        self.assertEqual(
            field_xso.var,
            self.field.var
        )

        self.assertEqual(
            field_xso.type_,
            self.field.FIELD_TYPE,
        )

        self.assertSequenceEqual(
            field_xso.values,
            []
        )

        self.assertEqual(
            field_xso.desc,
            "some description",
        )

        self.assertEqual(
            field_xso.label,
            "some label",
        )

        self.assertEqual(
            field_xso.required,
            unittest.mock.sentinel.required,
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )

    def test_render_uses_local_values(self):
        self.bf.desc = "some description"
        self.bf.label = "some label"
        self.bf.required = unittest.mock.sentinel.required

        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"

        field_xso = self.bf.render(
            use_local_metadata=False
        )

        self.assertIsInstance(
            field_xso,
            forms_xso.Field,
        )

        self.assertEqual(
            field_xso.var,
            self.field.var
        )

        self.assertEqual(
            field_xso.type_,
            self.field.FIELD_TYPE,
        )

        self.assertSequenceEqual(
            field_xso.values,
            []
        )

        self.assertEqual(
            field_xso.desc,
            "some description",
        )

        self.assertEqual(
            field_xso.label,
            "some label",
        )

        self.assertEqual(
            field_xso.required,
            unittest.mock.sentinel.required,
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )

    def test_render_inhibit_inheritance_of_metadata(self):
        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"

        field_xso = self.bf.render(
            use_local_metadata=False
        )

        self.assertIsInstance(
            field_xso,
            forms_xso.Field,
        )

        self.assertEqual(
            field_xso.var,
            self.field.var
        )

        self.assertEqual(
            field_xso.type_,
            self.field.FIELD_TYPE,
        )

        self.assertSequenceEqual(
            field_xso.values,
            []
        )

        self.assertIsNone(
            field_xso.desc,
        )

        self.assertIsNone(
            field_xso.label,
        )

        self.assertFalse(
            field_xso.required,
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )


class TestBoundSingleValueField(unittest.TestCase):
    def setUp(self):
        self.field = unittest.mock.Mock()
        self.instance = unittest.mock.sentinel.instance
        self.bf = fields.BoundSingleValueField(
            self.field,
            self.instance
        )

    def tearDown(self):
        del self.bf
        del self.instance
        del self.field

    def test_setting_value_uses_field_type(self):
        self.bf.value = unittest.mock.sentinel.value

        self.field.type_.coerce.assert_called_with(
            unittest.mock.sentinel.value,
        )

        self.assertEqual(
            self.bf.value,
            self.field.type_.coerce(),
        )

    def test_access_without_setting_calls_through_to_field(self):
        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertEqual(
            result,
            self.field.default(),
        )

    def test_deleting_reverts_to_initial_behaviour(self):
        self.bf.value = "foo"

        del self.bf.value
        del self.bf.value

        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertEqual(
            result,
            self.field.default(),
        )

    def test_load_loads_value_with_parse(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            values=[unittest.mock.sentinel.value],
        )

        self.bf.load(field_xso)

        self.field.type_.parse.assert_called_with(
            unittest.mock.sentinel.value,
        )

        self.assertEqual(
            self.bf.value,
            self.field.type_.parse(),
        )

    def test_load_defaults_if_no_values_present(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
        )

        self.bf.load(field_xso)

        self.field.default.assert_called_with()

        self.assertEqual(
            self.bf.value,
            self.field.default(),
        )

    def test_load_ignores_additional_values(self):
        field_xso = forms_xso.Field(
            type_=forms_xso.FieldType.TEXT_SINGLE,
            values=[1, 2, 3]
        )

        self.bf.load(field_xso)

    def test_load_calls_base_impl(self):
        field_xso = forms_xso.Field()

        with unittest.mock.patch.object(
                fields.BoundField, "load") as load:
            self.bf.load(field_xso)

        load.assert_called_with(field_xso)

    def test_render_creates_Field_xso(self):
        self.field.type_.coerce.return_value = unittest.mock.sentinel.coerced

        self.bf.value = unittest.mock.sentinel.value

        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"

        self.field.type_.format.return_value = "formatted value"

        with unittest.mock.patch.object(
                fields.BoundField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.field.type_.format.assert_called_with(
            unittest.mock.sentinel.coerced,
        )

        self.assertSequenceEqual(
            field_xso.values,
            ["formatted value"]
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )

    def test_render_defaults_to_backing_field(self):
        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"
        self.field.type_.format.return_value = "formatted value"

        with contextlib.ExitStack() as stack:
            render = stack.enter_context(unittest.mock.patch.object(
                fields.BoundField,
                "render"
            ))
            render.return_value = forms_xso.Field()

            default = stack.enter_context(unittest.mock.patch.object(
                self.field,
                "default",
            ))
            default.return_value = unittest.mock.sentinel.defaulted

            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_once_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        default.assert_called_once_with()

        self.field.type_.format.assert_called_with(
            unittest.mock.sentinel.defaulted,
        )

        self.assertSequenceEqual(
            field_xso.values,
            ["formatted value"]
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )

    def test_renders_without_value_if_default_is_None(self):
        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"
        self.field.type_.format.return_value = "formatted value"

        with contextlib.ExitStack() as stack:
            render = stack.enter_context(unittest.mock.patch.object(
                fields.BoundField,
                "render"
            ))
            render.return_value = forms_xso.Field()

            default = stack.enter_context(unittest.mock.patch.object(
                self.field,
                "default",
            ))
            default.return_value = None

            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_once_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        default.assert_called_once_with()

        self.field.type_.format.assert_not_called()

        self.assertSequenceEqual(
            field_xso.values,
            []
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )


class TestBoundMultiValueField(unittest.TestCase):
    def setUp(self):
        self.field = unittest.mock.Mock()
        self.field.default.return_value = [
            unittest.mock.sentinel.v1,
            unittest.mock.sentinel.v2,
            unittest.mock.sentinel.v2,
        ]
        self.instance = unittest.mock.sentinel.instance
        self.bf = fields.BoundMultiValueField(
            self.field,
            self.instance,
        )

    def tearDown(self):
        del self.bf
        del self.instance
        del self.field

    def test_is_bound_field(self):
        self.assertTrue(issubclass(
            fields.BoundMultiValueField,
            fields.BoundField
        ))

    def test_access_without_set_delegates_to_field_and_cast_to_tuple(self):
        self.field.type_.coerce.side_effect = generate_values("c")

        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertSequenceEqual(
            self.field.type_.coerce.mock_calls,
            [
                unittest.mock.call(unittest.mock.sentinel.v1),
                unittest.mock.call(unittest.mock.sentinel.v2),
                unittest.mock.call(unittest.mock.sentinel.v2),
            ]
        )

        self.assertSequenceEqual(
            result,
            [
                unittest.mock.sentinel.c0,
                unittest.mock.sentinel.c1,
                unittest.mock.sentinel.c2,
            ]
        )

        self.assertIsInstance(
            result,
            tuple,
        )

    def test_setting_field_coerces_and_casts_to_tuple(self):
        self.field.type_.coerce.side_effect = generate_values("c")

        self.bf.value = [1, 2, 3]

        self.assertSequenceEqual(
            self.field.type_.coerce.mock_calls,
            [
                unittest.mock.call(1),
                unittest.mock.call(2),
                unittest.mock.call(3),
            ]
        )

        self.assertSequenceEqual(
            self.bf.value,
            [
                unittest.mock.sentinel.c0,
                unittest.mock.sentinel.c1,
                unittest.mock.sentinel.c2,
            ]
        )

        self.assertIsInstance(
            self.bf.value,
            tuple,
        )

    def test_deleting_reverts_to_initial_behaviour(self):
        self.field.type_.coerce.side_effect = lambda x: x

        self.bf.value = ["foo", "bar"]

        del self.bf.value
        del self.bf.value

        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertSequenceEqual(
            result,
            self.field.default(),
        )

    def test_load_parses_values(self):
        values = ["10", "20", "30"]
        field_xso = forms_xso.Field(
            values=list(values),
        )

        self.field.type_.parse.side_effect = generate_values("parsed")

        self.bf.load(field_xso)

        self.assertSequenceEqual(
            self.field.type_.mock_calls,
            [
                unittest.mock.call.parse(v)
                for v in values
            ]
        )

        self.assertSequenceEqual(
            self.bf.value,
            [
                getattr(unittest.mock.sentinel, "parsed{}".format(i))
                for i in range(3)
            ]
        )

    def test_load_calls_base_impl(self):
        field_xso = forms_xso.Field()

        with unittest.mock.patch.object(
                fields.BoundField, "load") as load:
            self.bf.load(field_xso)

        load.assert_called_with(field_xso)

    def test_render_creates_Field_xso(self):
        def generate_coerced_values():
            i = 0
            while True:
                yield getattr(unittest.mock.sentinel, "coerced{}".format(i))
                i += 1

        def generate_formatted_values():
            i = 0
            while True:
                yield "value {}".format(i)
                i += 1

        self.field.type_.format.side_effect = generate_formatted_values()
        self.field.type_.coerce.side_effect = generate_coerced_values()

        self.bf.value = [
            1, 2, 3
        ]

        self.field.FIELD_TYPE = forms_xso.FieldType.TEXT_SINGLE
        self.field.var = "some var"

        with unittest.mock.patch.object(
                fields.BoundField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.assertSequenceEqual(
            self.field.type_.format.mock_calls,
            [
                unittest.mock.call(getattr(unittest.mock.sentinel,
                                           "coerced{}".format(i)))
                for i in range(3)
            ]
        )

        self.assertSequenceEqual(
            field_xso.values,
            ["value 0", "value 1", "value 2"]
        )

        self.assertSequenceEqual(
            field_xso.options,
            [],
        )


class TestBoundOptionsField(unittest.TestCase):
    def setUp(self):
        self.field = unittest.mock.Mock()
        self.field.options = collections.OrderedDict([
            ("foo", "The Foo"),
            ("bar", "The Bar"),
            ("baz", "The Baz"),
        ])
        self.instance = unittest.mock.sentinel.instance
        self.bf = fields.BoundOptionsField(
            self.field,
            self.instance,
        )

    def tearDown(self):
        del self.bf
        del self.instance
        del self.field

    def test_is_bound_field(self):
        self.assertTrue(issubclass(
            fields.BoundOptionsField,
            fields.BoundField
        ))

    def test_options_calls_through_to_field(self):
        self.assertEqual(
            self.bf.options,
            self.field.options,
        )

        self.field.options = unittest.mock.sentinel.options

        self.assertEqual(
            self.bf.options,
            self.field.options,
        )

    def test_setting_options_evaluates_into_ordered_dict_with_coerce(self):
        self.field.type_.coerce.side_effect = generate_values("coerced")

        options = (
            (0, "foo"),
            (1, "bar"),
            (2, "baz")
        )

        self.bf.options = options

        self.assertSequenceEqual(
            tuple(self.bf.options.keys()),
            [getattr(unittest.mock.sentinel, "coerced{}".format(i))
             for i, (k, v) in enumerate(options)]
        )

        self.assertDictEqual(
            self.bf.options,
            collections.OrderedDict([
                (getattr(unittest.mock.sentinel, "coerced{}".format(i)), v)
                for i, (k, v) in enumerate(options)
            ]),
        )

    def test_delete_resets_options_to_field(self):
        self.bf.options = []

        self.assertDictEqual(
            self.bf.options,
            {},
        )

        del self.bf.options
        del self.bf.options

        self.assertEqual(
            self.bf.options,
            self.field.options,
        )

    def test_load_loads_options_using_parse(self):
        options = (
            ("1", "First"),
            ("2", "Second"),
            ("3", "Third"),
        )

        field_xso = forms_xso.Field(
            options=options,
        )

        self.bf.load(field_xso)

        self.assertDictEqual(
            self.bf.options,
            collections.OrderedDict(options),
        )

        self.assertSequenceEqual(
            tuple(self.bf.options.keys()),
            [k for k, v in options]
        )

    def test_load_calls_base_impl(self):
        field_xso = forms_xso.Field()

        with unittest.mock.patch.object(
                fields.BoundField, "load") as load:
            self.bf.load(field_xso)

        load.assert_called_with(field_xso)

    def test_render_emits_base_options_using_format(self):
        self.field.type_.format.side_effect = str
        self.field.default.return_value = None

        with unittest.mock.patch.object(
                fields.BoundField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.assertSequenceEqual(
            self.field.type_.format.mock_calls,
            [
                unittest.mock.call(k)
                for k in self.field.options.keys()
            ]
        )

        self.assertDictEqual(
            field_xso.options,
            self.field.options
        )

    def test_render_emits_overridden_options_using_format(self):
        self.field.type_.coerce.side_effect = int
        self.field.type_.format.side_effect = str
        self.field.default.return_value = None

        self.bf.options = [(1, "foo")]

        with unittest.mock.patch.object(
                fields.BoundField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.field.type_.format.assert_called_with(
            1
        )

        self.assertDictEqual(
            field_xso.options,
            {"1": "foo"}
        )


class TestBoundSelectField(unittest.TestCase):
    def setUp(self):
        self.field = unittest.mock.Mock()
        self.field.options = collections.OrderedDict([
            ("foo", "The Foo"),
            ("bar", "The Bar"),
            ("baz", "The Baz"),
        ])
        self.instance = unittest.mock.sentinel.instance
        self.bf = fields.BoundSelectField(
            self.field,
            self.instance,
        )

    def tearDown(self):
        del self.bf
        del self.instance
        del self.field

    def test_is_bound_options_field(self):
        self.assertTrue(issubclass(
            fields.BoundSelectField,
            fields.BoundOptionsField
        ))

    def test_access_without_setting_calls_through_to_field(self):
        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertEqual(
            result,
            self.field.default(),
        )

    def test_setting_value_enforces_membership_in_options(self):
        with self.assertRaisesRegex(
                ValueError,
                r"not in field options: \('foo', 'bar', 'baz'\)"):
            self.bf.value = "fnord"

    def test_setting_value_with_correct_option_allowed(self):
        self.bf.value = "foo"
        self.assertEqual(
            self.bf.value,
            "foo",
        )

    def test_deleting_reverts_to_initial_behaviour(self):
        self.bf.value = "foo"

        del self.bf.value
        del self.bf.value

        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertEqual(
            result,
            self.field.default(),
        )

    def test_load_loads_value_with_parse(self):
        field_xso = forms_xso.Field(
            values=["x", "fnord", "1"],
        )

        self.bf.load(field_xso)

        self.field.type_.parse.assert_called_with("x")

        self.assertEqual(
            self.bf.value,
            self.field.type_.parse(),
        )

    def test_load_calls_base_impl(self):
        field_xso = forms_xso.Field()

        with unittest.mock.patch.object(
                fields.BoundOptionsField, "load") as load:
            self.bf.load(field_xso)

        load.assert_called_with(field_xso)

    def test_render_emits_no_value_if_None(self):
        self.field.default.return_value = None

        with unittest.mock.patch.object(
                fields.BoundOptionsField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.assertSequenceEqual(
            field_xso.values,
            [],
        )

    def test_render_emits_selected_value_with_format(self):
        self.field.options = {1: "foo"}
        self.field.type_.format.side_effect = str

        self.bf.value = 1

        with unittest.mock.patch.object(
                fields.BoundOptionsField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag,
        )

        self.field.type_.format.assert_called_with(1)

        self.assertSequenceEqual(
            field_xso.values,
            ["1"],
        )


class TestBoundMultiSelectField(unittest.TestCase):
    def setUp(self):
        self.field = unittest.mock.Mock()
        self.field.options = collections.OrderedDict([
            ("foo", "The Foo"),
            ("bar", "The Bar"),
            ("baz", "The Baz"),
        ])
        self.field.default.return_value = []
        self.instance = unittest.mock.sentinel.instance
        self.bf = fields.BoundMultiSelectField(
            self.field,
            self.instance,
        )

    def tearDown(self):
        del self.bf
        del self.instance
        del self.field

    def test_is_bound_options_field(self):
        self.assertTrue(issubclass(
            fields.BoundMultiSelectField,
            fields.BoundOptionsField
        ))

    def test_access_without_setting_calls_through_to_field(self):
        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertSetEqual(
            set(result),
            set(self.field.default()),
        )

    def test_setting_value_enforces_membership_in_options(self):
        with self.assertRaisesRegex(
                ValueError,
                r"not in field options: \('foo', 'bar', 'baz'\)"):
            self.bf.value = {"fnord"}

    def test_setting_value_with_correct_option_allowed(self):
        self.bf.value = ["foo", "bar"]
        self.assertSetEqual(
            self.bf.value,
            {"foo", "bar"},
        )

        self.assertIsInstance(
            self.bf.value,
            frozenset,
        )

    def test_deleting_reverts_to_initial_behaviour(self):
        self.bf.value = ["foo", "baz"]

        del self.bf.value
        del self.bf.value

        result = self.bf.value

        self.field.default.assert_called_with()

        self.assertSetEqual(
            set(result),
            set(self.field.default()),
        )

    def test_load_loads_values_with_parse(self):
        values = ["x", "fnord", "1"]

        self.field.type_.parse.side_effect = generate_values("parsed")

        field_xso = forms_xso.Field(
            values=values,
        )

        self.bf.load(field_xso)

        self.assertSequenceEqual(
            self.field.type_.parse.mock_calls,
            [
                unittest.mock.call("x"),
                unittest.mock.call("fnord"),
                unittest.mock.call("1"),
            ]
        )

        self.assertSetEqual(
            self.bf.value,
            set(itertools.islice(generate_values("parsed"), 3))
        )

    def test_load_calls_base_impl(self):
        field_xso = forms_xso.Field()

        with unittest.mock.patch.object(
                fields.BoundOptionsField, "load") as load:
            self.bf.load(field_xso)

        load.assert_called_with(field_xso)

    def test_render_emits_values_with_format(self):
        self.field.type_.format.side_effect = generate_values("formatted")

        self.bf.value = {"foo", "bar"}

        with unittest.mock.patch.object(
                fields.BoundOptionsField,
                "render") as render:
            render.return_value = forms_xso.Field()
            field_xso = self.bf.render(
                use_local_metadata=unittest.mock.sentinel.flag
            )

        render.assert_called_with(
            use_local_metadata=unittest.mock.sentinel.flag
        )

        self.assertSetEqual(
            {
                arg
                for _, (arg, ), _ in self.field.type_.format.mock_calls
            },
            {
                "foo", "bar"
            }
        )

        self.assertSetEqual(
            set(field_xso.values),
            {
                unittest.mock.sentinel.formatted0,
                unittest.mock.sentinel.formatted1,
            }
        )


class TestAbstractDescriptor(unittest.TestCase):
    def test_uses_abcmeta(self):
        self.assertIsInstance(
            fields.AbstractDescriptor,
            abc.ABCMeta,
        )

    def test_is_abstract(self):
        with self.assertRaisesRegex(TypeError, "descriptor_keys"):
            fields.AbstractDescriptor()


class TestAbstractField(unittest.TestCase):
    class FakeField(fields.AbstractField):
        FIELD_TYPE = forms_xso.FieldType.HIDDEN

        def __init__(self, var, type_=xso.String(), *args, **kwargs):
            super().__init__(var, type_, *args, **kwargs)

        def create_bound(self, for_instance):
            pass

        def default(self):
            pass

    def setUp(self):
        self.f = self.FakeField(unittest.mock.sentinel.var)
        self.instance = unittest.mock.Mock()
        self.instance._descriptor_data = {}

    def tearDown(self):
        del self.instance
        del self.f

    def test_init(self):
        f = self.FakeField(unittest.mock.sentinel.var)
        self.assertFalse(f.required)
        self.assertIsNone(f.desc)
        self.assertIsNone(f.label)
        self.assertEqual(f.var, unittest.mock.sentinel.var)

    def test_init_required(self):
        f = self.FakeField(unittest.mock.sentinel.var,
                           required=True)
        self.assertTrue(f.required)

    def test_init_desc(self):
        desc = "foobar"
        f = self.FakeField(unittest.mock.sentinel.var,
                           desc=desc)
        self.assertEqual(f.desc, desc)

    def test_init_label(self):
        f = self.FakeField(unittest.mock.sentinel.var,
                           label="foobar")
        self.assertEqual(f.label, "foobar")

    def test_init_rejects_desc_with_newlines(self):
        regex = r"desc must not contain newlines"

        with self.assertRaisesRegex(ValueError, regex):
            self.FakeField(unittest.mock.sentinel.var,
                           desc="foo\r")

        with self.assertRaisesRegex(ValueError, regex):
            self.FakeField(unittest.mock.sentinel.var,
                           desc="foo\nbar")

        with self.assertRaisesRegex(ValueError, regex):
            self.FakeField(unittest.mock.sentinel.var,
                           desc="foo\r\nbar")

    def test_setting_desc_rejects_desc_with_newlines(self):
        regex = r"desc must not contain newlines"

        with self.assertRaisesRegex(ValueError, regex):
            self.f.desc = "foo\r"

        with self.assertRaisesRegex(ValueError, regex):
            self.f.desc = "foo\nbar"

        with self.assertRaisesRegex(ValueError, regex):
            self.f.desc = "foo\r\nbar"

    def test_type_is_not_writable(self):
        t = self.f.type_
        with self.assertRaises(AttributeError):
            self.f.type_ = t

    def test_init_type(self):
        f = self.FakeField("var", unittest.mock.sentinel.type_)
        self.assertEqual(
            f.type_,
            unittest.mock.sentinel.type_
        )

    def test_deleting_desc_sets_it_to_None(self):
        f = self.FakeField(unittest.mock.sentinel.var, desc="foobar")
        del f.desc
        self.assertIsNone(f.desc)

    def test_descriptor_keys(self):
        self.assertSequenceEqual(
            list(self.f.descriptor_keys()),
            [
                (fields.descriptor_ns, unittest.mock.sentinel.var),
            ]
        )

    def test_var_is_not_settable(self):
        with self.assertRaises(AttributeError):
            self.f.var = "foo"

    def test_make_bound_creates_new_and_stores_at_instance(self):
        with unittest.mock.patch.object(
                self.f,
                "create_bound") as create_bound:
            result = self.f.make_bound(self.instance)

        create_bound.assert_called_with(self.instance)

        self.assertEqual(
            result,
            create_bound(),
        )

        self.assertDictEqual(
            self.instance._descriptor_data,
            {
                self.f: result,
            }
        )

    def test_make_bound_returns_field_stored_at_instance(self):
        self.instance._descriptor_data[self.f] = \
            unittest.mock.sentinel.existing

        with unittest.mock.patch.object(
                self.f,
                "create_bound") as create_bound:
            result = self.f.make_bound(self.instance)

        self.assertFalse(create_bound.mock_calls)

        self.assertEqual(
            result,
            unittest.mock.sentinel.existing,
        )

        self.assertDictEqual(
            self.instance._descriptor_data,
            {
                self.f: unittest.mock.sentinel.existing,
            }
        )

    def test___get___uses_make_bound_and_returns(self):
        with unittest.mock.patch.object(
                self.f,
                "make_bound") as make_bound:
            result = self.f.__get__(self.instance, type(self.instance))

        make_bound.assert_called_with(self.instance)

        self.assertEqual(
            result,
            make_bound(),
        )

    def test___get___returns_self_for_None_instance(self):
        result = self.f.__get__(None, type(self.instance))

        self.assertIs(
            result,
            self.f,
        )


class TestTextSingle(unittest.TestCase):
    def setUp(self):
        self.f = fields.TextSingle(var="muc#foobar")

    def tearDown(self):
        del self.f

    def test_is_abstract_field(self):
        self.assertTrue(issubclass(
            fields.TextSingle,
            fields.AbstractField,
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.TextSingle.FIELD_TYPE,
            forms_xso.FieldType.TEXT_SINGLE,
        )

    def test_init(self):
        t = object()
        f = fields.TextSingle(var="muc#foobar", type_=t)
        self.assertEqual(f.var, "muc#foobar")
        self.assertIs(f.type_, t)

        f = fields.TextSingle("muc#foobar")
        self.assertEqual(f.var, "muc#foobar")

        with self.assertRaisesRegex(TypeError, "argument"):
            fields.TextSingle()

    def test_init_passes_to_abstract_field(self):
        f = fields.TextSingle(
            var="muc#foobar",
            required=True,
            desc="foobar"
        )
        self.assertTrue(f.required)
        self.assertEqual(f.desc, "foobar")

    def test_creates_single_value_bound(self):
        mock = instance_mock()
        with contextlib.ExitStack() as stack:
            BoundSingleValueField = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.forms.fields.BoundSingleValueField"
                )
            )

            result = self.f.create_bound(mock)

        BoundSingleValueField.assert_called_with(
            self.f,
            mock,
        )

        self.assertEqual(
            result,
            BoundSingleValueField(),
        )


class TestJIDSingle(unittest.TestCase):
    def setUp(self):
        self.f = fields.JIDSingle(var="muc#foobar")

    def tearDown(self):
        del self.f

    def test_is_abstract_field(self):
        self.assertTrue(issubclass(
            fields.JIDSingle,
            fields.AbstractField
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.JIDSingle.FIELD_TYPE,
            forms_xso.FieldType.JID_SINGLE,
        )

    def test_type_is_JID(self):
        f = fields.JIDSingle("foo")
        self.assertIsInstance(f.type_, xso.JID)

    def test_init_still_rejects_type__argument(self):
        t = object()
        with self.assertRaisesRegex(TypeError, "type_"):
            fields.JIDSingle("foo", type_=t)

    def test_init_passes_to_abstract_field(self):
        f = fields.JIDSingle(
            var="muc#foobar",
            required=True,
            desc="foobar"
        )
        self.assertTrue(f.required)
        self.assertEqual(f.desc, "foobar")

    def test_init_default(self):
        self.assertIsNone(self.f.default())

        f = fields.JIDSingle("foo", default=TEST_JID)

        self.assertEqual(
            f.default(),
            TEST_JID,
        )

    def test_creates_single_value_bound(self):
        mock = instance_mock()
        with contextlib.ExitStack() as stack:
            BoundSingleValueField = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.forms.fields.BoundSingleValueField"
                )
            )

            result = self.f.create_bound(mock)

        BoundSingleValueField.assert_called_with(
            self.f,
            mock,
        )

        self.assertEqual(
            result,
            BoundSingleValueField(),
        )


class TestBoolean(unittest.TestCase):
    def setUp(self):
        self.f = fields.Boolean(var="muc#foobar")

    def tearDown(self):
        del self.f

    def test_is_abstract_field(self):
        self.assertTrue(issubclass(
            fields.Boolean,
            fields.AbstractField
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.Boolean.FIELD_TYPE,
            forms_xso.FieldType.BOOLEAN,
        )

    def test_type_is_Bool(self):
        f = fields.Boolean("foo")
        self.assertIsInstance(f.type_, xso.Bool)

    def test_init_rejects_type__argument(self):
        t = object()
        with self.assertRaisesRegex(TypeError, "type_"):
            fields.Boolean("foo", type_=t)

    def test_init_passes_to_abstract_field(self):
        f = fields.Boolean(
            var="muc#foobar",
            required=True,
            desc="foobar"
        )
        self.assertTrue(f.required)
        self.assertEqual(f.desc, "foobar")

    def test_default_default(self):
        self.assertEqual(
            fields.Boolean("foo").default(),
            False,
        )

    def test_creates_single_value_bound(self):
        mock = instance_mock()
        with contextlib.ExitStack() as stack:
            BoundSingleValueField = stack.enter_context(
                unittest.mock.patch(
                    "aioxmpp.forms.fields.BoundSingleValueField"
                )
            )

            result = self.f.create_bound(mock)

        BoundSingleValueField.assert_called_with(
            self.f,
            mock,
        )

        self.assertEqual(
            result,
            BoundSingleValueField(),
        )


class TestTextPrivate(unittest.TestCase):
    def test_is_text_single(self):
        self.assertTrue(issubclass(
            fields.TextPrivate,
            fields.TextSingle
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.TextPrivate.FIELD_TYPE,
            forms_xso.FieldType.TEXT_PRIVATE,
        )

    def test_init_still_accepts_type__argument(self):
        t = object()
        f = fields.TextPrivate("foo", type_=t)
        self.assertIs(f.type_, t)

    def test_init_passes_to_abstract_field(self):
        f = fields.TextPrivate(
            var="muc#foobar",
            required=True,
            desc="foobar"
        )
        self.assertTrue(f.required)
        self.assertEqual(f.desc, "foobar")


class TestTextMulti(unittest.TestCase):
    def setUp(self):
        self.f = fields.TextMulti("foo")

    def tearDown(self):
        del self.f

    def test_field_type(self):
        self.assertEqual(
            fields.TextMulti.FIELD_TYPE,
            forms_xso.FieldType.TEXT_MULTI,
        )

    def test_is_abstract_field(self):
        self.assertTrue(
            issubclass(
                fields.TextMulti,
                fields.AbstractField,
            )
        )

    def test_init(self):
        t = object()
        f = fields.TextMulti(var="muc#foobar", type_=t)
        self.assertEqual(f.var, "muc#foobar")
        self.assertIs(f.type_, t)

        f = fields.TextMulti("muc#foobar")
        self.assertEqual(f.var, "muc#foobar")
        self.assertIsInstance(f.type_, xso.String)

        with self.assertRaisesRegex(TypeError, "argument"):
            fields.TextMulti()

    def test_init_passes_to_abstract_field(self):
        f = fields.TextMulti(
            var="muc#foobar",
            required=True,
            desc="foobar"
        )
        self.assertTrue(f.required)
        self.assertEqual(f.desc, "foobar")

    def test_type_is_not_writable(self):
        with self.assertRaises(AttributeError):
            self.f.type_ = self.f.type_

    def test_init_default(self):
        f = fields.TextMulti(
            var="muc#foobar",
            default=unittest.mock.sentinel.default
        )
        self.assertEqual(
            f.default(),
            unittest.mock.sentinel.default
        )

    def test_init_default_default(self):
        f = fields.TextMulti(
            var="muc#foobar",
        )
        self.assertEqual(
            f.default(),
            ()
        )

    def test_create_bound_creates_multi_value_field(self):
        with contextlib.ExitStack() as stack:
            BoundMultiValueField = stack.enter_context(unittest.mock.patch(
                "aioxmpp.forms.fields.BoundMultiValueField"
            ))

            result = self.f.create_bound(
                unittest.mock.sentinel.instance
            )

        BoundMultiValueField.assert_called_with(
            self.f,
            unittest.mock.sentinel.instance,
        )

        self.assertEqual(
            result,
            BoundMultiValueField(),
        )


class TestJIDMulti(unittest.TestCase):
    def setUp(self):
        self.f = fields.JIDMulti("foo")

    def tearDown(self):
        del self.f

    def test_is_abstract_field(self):
        self.assertTrue(
            issubclass(
                fields.JIDMulti,
                fields.AbstractField,
            )
        )

    def test_type_(self):
        self.assertIsInstance(
            self.f.type_,
            xso.JID,
        )

    def test_field_type(self):
        self.assertEqual(
            fields.JIDMulti.FIELD_TYPE,
            forms_xso.FieldType.JID_MULTI,
        )

    def test_create_bound_creates_multi_value_field(self):
        with contextlib.ExitStack() as stack:
            BoundMultiValueField = stack.enter_context(unittest.mock.patch(
                "aioxmpp.forms.fields.BoundMultiValueField"
            ))

            result = self.f.create_bound(
                unittest.mock.sentinel.instance
            )

        BoundMultiValueField.assert_called_with(
            self.f,
            unittest.mock.sentinel.instance,
        )

        self.assertEqual(
            result,
            BoundMultiValueField(),
        )


class TestAbstractChoiceField(unittest.TestCase):
    class FakeChoiceField(fields.AbstractChoiceField):
        FIELD_TYPE = forms_xso.FieldType.HIDDEN

        def create_bound(self, for_instance):
            pass

        def default(self):
            pass

    def setUp(self):
        self.f = self.FakeChoiceField("var")

    def tearDown(self):
        del self.f

    def test_is_abstract_field(self):
        self.assertTrue(issubclass(
            fields.AbstractChoiceField,
            fields.AbstractField,
        ))

    def test_is_abstract(self):
        with self.assertRaisesRegex(
                TypeError,
                r"abstract.* create_bound, default"):
            fields.AbstractChoiceField()

    def test_init_options_with_coerce(self):
        type_ = unittest.mock.Mock()
        type_.coerce.side_effect = generate_values("coerced")

        f = self.FakeChoiceField(
            "var",
            options=[
                ("1", "foo"),
                ("2", "bar"),
                ("3", "baz"),
            ],
            type_=type_
        )

        self.assertSequenceEqual(
            type_.coerce.mock_calls,
            [
                unittest.mock.call("1"),
                unittest.mock.call("2"),
                unittest.mock.call("3"),
            ]
        )

        self.assertDictEqual(
            f.options,
            {
                unittest.mock.sentinel.coerced0: "foo",
                unittest.mock.sentinel.coerced1: "bar",
                unittest.mock.sentinel.coerced2: "baz",
            }
        )

        self.assertSequenceEqual(
            tuple(f.options.keys()),
            list(itertools.islice(generate_values("coerced"), 3))
        )

        self.assertIsInstance(
            f.options,
            collections.OrderedDict,
        )

    def test_init_default_default(self):
        self.assertIsNone(self.f.default())

    def test_init_type(self):
        t = object()
        f = self.FakeChoiceField("foo", type_=t)
        self.assertIs(f.type_, t)

    def test_type_is_not_settable(self):
        with self.assertRaises(AttributeError):
            self.f.type_ = xso.JID()


class TestListSingle(unittest.TestCase):
    def setUp(self):
        self.f = fields.ListSingle(
            "var",
            options=[
                ("foo", "The Foo"),
                ("bar", "The Bar"),
                ("baz", "The Baz"),
            ]
        )

    def tearDown(self):
        del self.f

    def test_is_abstract_choice_field(self):
        self.assertTrue(issubclass(
            fields.ListSingle,
            fields.AbstractChoiceField
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.ListSingle.FIELD_TYPE,
            forms_xso.FieldType.LIST_SINGLE
        )

    def test_init_default_default(self):
        self.assertIsNone(self.f.default())

    def test_init_default_rejects_not_None_default_not_in_options(self):
        with self.assertRaisesRegex(ValueError,
                                    r"invalid default: not in options"):
            fields.ListSingle(
                "var",
                options=[("foo", "The Foo")],
                default="bar",
            )

    def test_init_default(self):
        f = fields.ListSingle(
            "var",
            options=[("foo", "The Foo")],
            default="foo",
        )

        self.assertEqual(
            f.default(),
            "foo",
        )

    def test_create_bound_creates_BoundSelectField(self):
        with contextlib.ExitStack() as stack:
            BoundSelectField = stack.enter_context(
                unittest.mock.patch("aioxmpp.forms.fields.BoundSelectField")
            )

            result = self.f.create_bound(unittest.mock.sentinel.instance)

        BoundSelectField.assert_called_with(
            self.f,
            unittest.mock.sentinel.instance,
        )

        self.assertEqual(
            result,
            BoundSelectField(),
        )


class TestListMulti(unittest.TestCase):
    def setUp(self):
        self.f = fields.ListMulti(
            "var",
            options=[
                ("foo", "The Foo"),
                ("bar", "The Bar"),
                ("baz", "The Baz"),
            ]
        )

    def tearDown(self):
        del self.f

    def test_is_abstract_choice_field(self):
        self.assertTrue(issubclass(
            fields.ListMulti,
            fields.AbstractChoiceField
        ))

    def test_field_type(self):
        self.assertEqual(
            fields.ListMulti.FIELD_TYPE,
            forms_xso.FieldType.LIST_MULTI
        )

    def test_init_default_default(self):
        self.assertSetEqual(
            self.f.default(),
            set(),
        )

    def test_init_default(self):
        f = fields.ListMulti(
            "var",
            options=self.f.options,
            default={"foo", "baz"}
        )

        self.assertSetEqual(
            f.default(),
            {"foo", "baz"}
        )

        self.assertIsInstance(
            f.default(),
            frozenset,
        )

    def test_init_rejects_default_with_values_not_in_options(self):
        with self.assertRaisesRegex(ValueError,
                                    r"invalid default: not in options"):
            fields.ListMulti(
                "var",
                options=[("foo", "The Foo")],
                default={"foo", "bar"},
            )

    def test_create_bound_creates_BoundMultiSelectField(self):
        with contextlib.ExitStack() as stack:
            BoundMultiSelectField = stack.enter_context(
                unittest.mock.patch("aioxmpp.forms.fields.BoundMultiSelectField")
            )

            result = self.f.create_bound(unittest.mock.sentinel.instance)

        BoundMultiSelectField.assert_called_with(
            self.f,
            unittest.mock.sentinel.instance,
        )

        self.assertEqual(
            result,
            BoundMultiSelectField(),
        )



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp
import aioxmpp.mdr
import aioxmpp.tracking

from aioxmpp.e2etest import (
    blocking_timed,
    TestCase,
)


LANG = aioxmpp.structs.LanguageTag.fromstr("en-gb")


class TestMessaging(TestCase):
    @blocking_timed
    @asyncio.coroutine
    def setUp(self):
        self.a, self.b = yield from asyncio.gather(
            self.provisioner.get_connected_client(),
            self.provisioner.get_connected_client(),
        )
        self.a_mdr = self.a.summon(aioxmpp.DeliveryReceiptsService)
        self.b_mdr = self.b.summon(aioxmpp.DeliveryReceiptsService)
        self.b_recv = self.b.summon(
            aioxmpp.dispatcher.SimpleMessageDispatcher
        )

    @blocking_timed
    @asyncio.coroutine
    def test_send_with_attached_tracker(self):
        msg_delivered = asyncio.Event()

        def state_change_cb(new_state, *args):
            if (new_state ==
                    aioxmpp.tracking.MessageState.DELIVERED_TO_RECIPIENT):
                msg_delivered.set()

        msg = aioxmpp.Message(type_=aioxmpp.MessageType.CHAT,
                              to=self.b.local_jid)
        msg.body[LANG] = "This is a non-colourful test message!"

        tracker = self.a_mdr.attach_tracker(msg)
        tracker.on_state_changed.connect(state_change_cb)

        msg_recvd = asyncio.Future()

        def cb(message):
            msg_recvd.set_result(message)

        self.b_recv.register_callback(aioxmpp.MessageType.CHAT, None, cb)

        yield from self.a.send(msg)

        msg_b = yield from msg_recvd
        self.assertDictEqual(
            msg_b.body,
            {
                LANG: "This is a non-colourful test message!"
            }
        )
        self.assertTrue(msg_b.xep0184_request_receipt)

        self.assertFalse(msg_delivered.is_set())

        response = aioxmpp.mdr.compose_receipt(msg_b)
        yield from self.b.send(response)

        yield from msg_delivered.wait()



########################################################################
# File name: test_xmltestutils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

from aioxmpp.utils import etree

from aioxmpp.xmltestutils import XMLTestCase, element_path


class TestTestUtils(unittest.TestCase):
    def test_element_path(self):
        el = etree.fromstring("<foo><bar><baz /></bar>"
                              "<subfoo />"
                              "<bar><baz /></bar></foo>")
        baz1 = el[0][0]
        subfoo = el[1]
        baz2 = el[2][0]

        self.assertEqual(
            "/foo",
            element_path(el))
        self.assertEqual(
            "/foo/bar[0]/baz[0]",
            element_path(baz1))
        self.assertEqual(
            "/foo/subfoo[0]",
            element_path(subfoo))
        self.assertEqual(
            "/foo/bar[1]/baz[0]",
            element_path(baz2))


class TestXMLTestCase(XMLTestCase):
    def test_assertSubtreeEqual_tag(self):
        t1 = etree.fromstring("<foo />")
        t2 = etree.fromstring("<bar />")

        with self.assertRaisesRegex(AssertionError, "tag mismatch"):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_attr_key_missing(self):
        t1 = etree.fromstring("<foo a='1'/>")
        t2 = etree.fromstring("<foo />")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2, ignore_surplus_attr=True)

    def test_assertSubtreeEqual_attr_surplus_key(self):
        t1 = etree.fromstring("<foo a='1'/>")
        t2 = etree.fromstring("<foo />")
        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_attr_allow_surplus(self):
        t1 = etree.fromstring("<foo />")
        t2 = etree.fromstring("<foo a='1'/>")
        self.assertSubtreeEqual(t1, t2, ignore_surplus_attr=True)

    def test_assertSubtreeEqual_attr_value_mismatch(self):
        t1 = etree.fromstring("<foo a='1'/>")
        t2 = etree.fromstring("<foo a='2'/>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_attr_value_mismatch_allow_surplus(self):
        t1 = etree.fromstring("<foo a='1'/>")
        t2 = etree.fromstring("<foo a='1' b='2'/>")

        self.assertSubtreeEqual(t1, t2, ignore_surplus_attr=True)

    def test_assertSubtreeEqual_missing_child(self):
        t1 = etree.fromstring("<foo><bar/></foo>")
        t2 = etree.fromstring("<foo />")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_surplus_child(self):
        t1 = etree.fromstring("<foo><bar/></foo>")
        t2 = etree.fromstring("<foo><bar/><bar/></foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_allow_surplus_child(self):
        t1 = etree.fromstring("<foo />")
        t2 = etree.fromstring("<foo><bar/></foo>")

        self.assertSubtreeEqual(t1, t2, ignore_surplus_children=True)

        t1 = etree.fromstring("<foo><bar/></foo>")
        t2 = etree.fromstring("<foo><bar/><bar/><fnord /></foo>")

        self.assertSubtreeEqual(t1, t2, ignore_surplus_children=True)

    def test_assertSubtreeEqual_allow_relative_reordering(self):
        t1 = etree.fromstring("<foo><bar/><baz/></foo>")
        t2 = etree.fromstring("<foo><baz/><bar/></foo>")

        self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_forbid_reordering_of_same(self):
        t1 = etree.fromstring("<foo><bar a='1' /><bar a='2' /></foo>")
        t2 = etree.fromstring("<foo><bar a='2' /><bar a='1' /></foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_strict_ordering(self):
        t1 = etree.fromstring("<foo><bar/><baz/></foo>")
        t2 = etree.fromstring("<foo><baz/><bar/></foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2, strict_ordering=True)

    def test_assertSubtreeEqual_text(self):
        t1 = etree.fromstring("<foo>text1</foo>")
        t2 = etree.fromstring("<foo>text2</foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_text_including_tails(self):
        t1 = etree.fromstring("<foo>t<a/>ext1</foo>")
        t2 = etree.fromstring("<foo>te<a/>xt1</foo>")

        self.assertSubtreeEqual(t1, t2)

    def test_assertSubtreeEqual_text_no_join_text_parts(self):
        t1 = etree.fromstring("<foo>t<a/>ext1</foo>")
        t2 = etree.fromstring("<foo>te<a/>xt1</foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2, join_text_parts=False)

    def test_assertSubtreeEqual_text_no_join_text_parts_is_strict(self):
        t1 = etree.fromstring("<foo><a/>text1</foo>")
        t2 = etree.fromstring("<foo>text1<a/></foo>")

        with self.assertRaises(AssertionError):
            self.assertSubtreeEqual(t1, t2, join_text_parts=False)



########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

from aioxmpp.utils import namespaces

import aioxmpp.ping

from aioxmpp.e2etest import (
    TestCase,
    blocking,
    blocking_timed,
)


class TestPing(TestCase):
    @blocking
    @asyncio.coroutine
    def setUp(self):
        self.source, self.unimplemented, self.implemented = \
            yield from asyncio.gather(
                self.provisioner.get_connected_client(
                    services=[aioxmpp.ping.PingService],
                ),
                self.provisioner.get_connected_client(
                    services=[aioxmpp.DiscoClient],
                ),
                self.provisioner.get_connected_client(
                    services=[aioxmpp.ping.PingService],
                ),
            )

    @blocking_timed
    @asyncio.coroutine
    def test_ping_raises_error_condition(self):
        ping_svc = self.source.summon(aioxmpp.ping.PingService)

        with self.assertRaisesRegexp(aioxmpp.XMPPCancelError,
                                     "service-unavailable"):
            yield from ping_svc.ping(self.unimplemented.local_jid)

    @blocking_timed
    @asyncio.coroutine
    def test_ping_server(self):
        ping_svc = self.source.summon(aioxmpp.ping.PingService)

        yield from ping_svc.ping(self.unimplemented.local_jid.replace(
            localpart=None,
            resource=None,
        ))

    @blocking_timed
    @asyncio.coroutine
    def test_ping_works_with_peer_with_ping_implementation(self):
        ping_svc = self.source.summon(aioxmpp.ping.PingService)

        self.assertIsNone(
            (yield from ping_svc.ping(self.implemented.local_jid))
        )

    @blocking_timed
    @asyncio.coroutine
    def test_ping_service_exports_feature(self):
        info = yield from self.unimplemented.summon(
            aioxmpp.DiscoClient
        ).query_info(
            self.source.local_jid,
        )

        self.assertIn(
            namespaces.xep0199_ping,
            info.features,
        )

    @blocking_timed
    @asyncio.coroutine
    def test_ping_service_replies_to_ping(self):
        req = aioxmpp.IQ(
            type_=aioxmpp.IQType.GET,
            to=self.source.local_jid,
            payload=aioxmpp.ping.Ping(),
        )

        resp = yield from self.unimplemented.send(req)

        self.assertIsInstance(resp, aioxmpp.ping.Ping)
